'''Benchmarks various aspects of pySubnetSB.'''

"""
Evaluates subset detection by seeing if the reference network can be found when it is combined
with a filer network.

Key data structures:
    BenchmarkResult is a dataframe
        Index: index of the network used
        Columns:
            time - execution time
            num_permutations - number of permutations
"""

import scipy.special  # type: ignore
from pySubnetSB.significance_calculator_core import SignificanceCalculatorCore  # type: ignore
from pySubnetSB.reaction_constraint import ReactionConstraint  # type: ignore
from pySubnetSB.constraint_option_collection import SpeciesConstraintOptionCollection  # type: ignore
from pySubnetSB.constraint_option_collection import ReactionConstraintOptionCollection  # type: ignore
from pySubnetSB.species_constraint import SpeciesConstraint    # type: ignore
from pySubnetSB.network import Network      # type: ignore
import pySubnetSB.constants as cn # type: ignore

import collections
import matplotlib.pyplot as plt # type: ignore
import math
import numpy as np
import pandas as pd  # type: ignore
import matplotlib.pyplot as plt
import scipy
import seaborn as sns  # type: ignore
import tqdm  # type: ignore
import time
from typing import List, Optional, Tuple

NULL_DF = pd.DataFrame()
C_TIME = 'time'
C_LOG10_NUM_PERMUTATION = 'log10num_permutation'
C_NUM_REFERENCE = 'num_reference'
C_NUM_TARGET = 'num_target'
TITLE_FONT_SIZE_INCREMENT = 2
POC_SIGNIFICANCE = "poc_significance"

DimensionResult = collections.namedtuple('DimensionResult', ['dataframe', 'short_to_long_dct'])
EvaluateConstraintsResult = collections.namedtuple('EvaluateConstraintsResult',
      ['reference_size', 'target_size', 'species_dimension_result', 'reaction_dimension_result'])


# A study result is a container of the results of multiple benchmarks
StudyResult = collections.namedtuple('StudyResult', ['is_categorical_id', 'study_ids', 'benchmark_results'])
#  is_categorical_id: bool # True if study_ids are categorical
#  study_ids: List[str | float]
#  benchmark_results: List[pd.DataFrame]


class Benchmark(object):
    def __init__(self, reference_size:int, fill_size:int=0, num_iteration:int=1000,
          is_contains_reference:bool=True)->None:
        """
        Args:
            reference_size (int): size of the reference network (species, reaction)
            fill_size (int): size of the filler network (species, reaction) used in subsets
            is_contains_reference (bool, Optional): target network contains reference network. Defaults to True.
        """
        self.num_reaction = reference_size
        self.num_species = reference_size
        self.num_iteration = num_iteration
        self.fill_size = fill_size
        # Calculated
        self.reference_networks = [Network.makeRandomNetworkByReactionType(self.num_reaction, self.num_species)
                for _ in range(num_iteration)]
        if fill_size > 0:
            self.target_networks = [n.fill(num_fill_reaction=self.num_reaction,
              num_fill_species=self.num_species) for n in self.reference_networks]
        if is_contains_reference:
            self.target_networks = [n.fill(num_fill_reaction=self.fill_size,
                num_fill_species=self.fill_size) for n in self.reference_networks]
        else:
            target_size = reference_size + self.fill_size
            self.reference_networks = [Network.makeRandomNetworkByReactionType(target_size, target_size)
                for _ in range(num_iteration)]
        self.benchmark_result_df = NULL_DF  # Updated with result of run

    @staticmethod 
    def _getConstraintClass(is_species:bool=True):
        if is_species:
            return SpeciesConstraint
        else:
            return ReactionConstraint

    def run(self, is_subnet:bool=True, is_species:bool=True)->pd.DataFrame:
        """Evaluates the effectiveness of reaction and species constraints.

        Args:
            is_subnet (bool, optional): is subset constraint
            is_species (bool, optional): is species constraint

        Returns:
            pd.DataFrame: _description_
        """
        
        times:list = []
        num_permutations:list = []
        constraint_cls = self._getConstraintClass(is_species)
        for reference_network, target_network in zip(self.reference_networks, self.target_networks):
            # Species
            start = time.time()
            reference_constraint = constraint_cls(
                    reference_network.reactant_nmat, reference_network.product_nmat,
                    is_subnet=is_subnet)
            if is_subnet:
                target_constraint = constraint_cls(
                    target_network.reactant_nmat, target_network.product_nmat, is_subnet=is_subnet)
            else:
                new_target_network, _ = reference_network.permute()
                target_constraint = constraint_cls(
                    new_target_network.reactant_nmat, new_target_network.product_nmat,
                    is_subnet=is_subnet)
            compatibility_collection = reference_constraint.makeCompatibilityCollection(
                  target_constraint).compatibility_collection
            times.append(time.time() - start)
            num_permutations.append(compatibility_collection.log10_num_assignment)
        self.benchmark_result_df = pd.DataFrame({C_TIME: times, C_LOG10_NUM_PERMUTATION: num_permutations})
        return self.benchmark_result_df

    @classmethod 
    def plotConstraintStudy(cls, reference_size:int, fill_size:int, num_iteration:int,
          is_plot:bool=True, **kwargs):
        """Plot the results of a study of constraints.

        Args:
            reference_size (int): size of the reference network (species, reaction)
            fill_size (int): size of the filler network (species, reaction) used in subsets
            num_iteration (int): number of iterations
            is_plot (bool, optional): Plot the results. Defaults to True.
            kwargs: constructor options
        """
        benchmark = Benchmark(reference_size, fill_size=fill_size, num_iteration=num_iteration)
        def doPlot(df:pd.DataFrame, node_str:str, is_subnet:bool=False, pos:int=0):
            ax = axes.flatten()[pos]
            xv = np.array(range(len(df)))
            yv = df[C_LOG10_NUM_PERMUTATION].values.copy()
            sel = yv == 0
            yv[sel] = 1
            yv = yv
            yv = np.sort(yv)
            xv = xv/len(yv)
            ax.plot(xv, yv)
            title = 'Subset' if is_subnet else 'Full'
            title = node_str + ' ' + title
            ax.set_title(title)
            ax.set_ylim(0, 10)
            ax.set_xlim(0, 1)
            if pos in [0, 3]:
                ax.set_ylabel('Log10 permutations')
            else:
                ax.set_yticks([])
            if pos in [3, 4, 5]:
                ax.set_xlabel('Culmulative networks')
            else:
                ax.set_xticks([])
            ax.plot([0, 1], [8, 8], 'k--')
    #####
        fig, axes = plt.subplots(2, 3)
        suptitle = f"CDF of permutations: reference_size={benchmark.num_reaction}; fill_size={benchmark.fill_size}"
        fig.suptitle(suptitle)
        # Collect data and construct plots
        pos = 0
        for is_subnet in [False, True]:
            dataframe_dct:dict = {}
            for is_species in [False, True]:
                if is_species:
                    node_str = 'Spc.'
                else:
                    node_str = 'Rct.'
                dataframe_dct[is_species] = benchmark.run(is_species=is_species, is_subnet=is_subnet)
                doPlot(dataframe_dct[is_species], node_str, is_subnet, pos=pos)
                pos += 1
            # Construct totals plot
            df = dataframe_dct[True].copy()
            df += dataframe_dct[False]
            doPlot(df, "Total", is_subnet, pos=pos)
            pos += 1
        if is_plot:
            plt.show()

    @classmethod
    def plotHeatmap(cls, num_references:List[int], num_targets:List[int], percentile:int=50,
          num_iteration:int=20, is_contains_reference=True, is_plot:bool=True,
          is_no_constraint:bool=False, title:Optional[str]=None,
          ax=None, font_size:int=8, is_cbar:bool=True, num_digit:int=1)->plt.Axes:
        """Plot a heatmap of the log10 of number of permutations.

        Args:
            num_references (List[int]): number of reference networks
            num_targets (List[int]): number of target networks
            percentile (int): percentile of distribution of the log number of permutation
            is_plot (bool, optional): plot the heatmap. Defaults to True.
            is_not_constraint (bool, optional): True if no constraints are applied. Defaults to False.  
            title (str, optional): title of the plot. Defaults to None.
            ax: Matplotlib axes
            font_size (int, optional): font size of the color bar. Defaults to 8.
            is_cbar (bool, optional): show the color bar. Defaults to True.

        Returns:
            matplotlib.axes._axes.Axes: _description_
        """
        #####
        def _round(value:float, num_digit:int)->float:
            if np.isnan(value):
                return np.nan
            if num_digit == 0:
                result = round(value)
            else:
                result = np.round(value, num_digit)
            return result
        #####
        def calculate(is_species:bool)->pd.DataFrame:
            df = benchmark.run(is_species=is_species, is_subnet=is_subnet)
            df[C_LOG10_NUM_PERMUTATION] = [max(v, 0) for v in df[C_LOG10_NUM_PERMUTATION]]
            return df
        #####
        # Construct the dataj
        C_LOG10_P10 = 'log10_num_permutation_p10'
        C_LOG10_P90 = 'log10_num_permutation_p90'
        C_LABEL = 'label'
        data_dct:dict = {C_NUM_REFERENCE: [], C_NUM_TARGET: [], C_LOG10_NUM_PERMUTATION: [],
              C_LOG10_P10: [], C_LOG10_P90: [], C_LABEL: []}
        for reference_size in num_references:
            for target_size in num_targets:
                fill_size = target_size - reference_size
                data_dct[C_NUM_REFERENCE].append(reference_size)
                data_dct[C_NUM_TARGET].append(target_size)
                is_subnet = fill_size > 0
                if fill_size < 0:
                    result = np.nan
                    percentile_10 = np.nan
                    percentile_90 = np.nan
                else:
                    fill_size = max(1, fill_size)
                    if is_no_constraint:
                        result = 2*cls.calculateNoconstraintLog10NumAssignment(reference_size, target_size)
                        percentile_10 = result
                        percentile_90 = result
                    else:
                        benchmark = cls(reference_size, fill_size=fill_size,
                            is_contains_reference=is_contains_reference, num_iteration=num_iteration)
                        df_species = calculate(is_species=True)
                        df_reaction = calculate(is_species=False)
                        df = df_species + df_reaction
                        result = _round(np.percentile(df[C_LOG10_NUM_PERMUTATION].values, percentile),
                                num_digit)
                        # Calculate multiple percentiles
                        percentile_10 = _round(np.percentile(df[C_LOG10_NUM_PERMUTATION].values, 10),
                              num_digit)
                        percentile_90 = _round(np.percentile(df[C_LOG10_NUM_PERMUTATION].values, 90),
                              num_digit)
                data_dct[C_LOG10_NUM_PERMUTATION].append(_round(result, num_digit))
                if np.isnan(percentile_10):
                    data_dct[C_LABEL].append("")
                else:
                    if is_no_constraint:
                        data_dct[C_LABEL].append(f"{_round(result, num_digit)}")
                    else:
                        annot = f"{_round(result, 1)}\n({np.round(percentile_10, num_digit)}, "
                        annot += f"{_round(percentile_90, num_digit)})"
                        data_dct[C_LABEL].append(annot)
                data_dct[C_LOG10_P10].append(_round(percentile_10, num_digit))
                data_dct[C_LOG10_P90].append(_round(percentile_90, num_digit))
        # Construct the dataframe
        df = pd.DataFrame(data_dct)
        df = df.rename(columns={C_NUM_REFERENCE: 'Reference', C_NUM_TARGET: 'Target'})
        if percentile > 0:
            df[C_LOG10_NUM_PERMUTATION] = [_round(v, num_digit) for v in df[C_LOG10_NUM_PERMUTATION]]
        else:
            df[C_LOG10_NUM_PERMUTATION] = [_round(v, num_digit)
                  for v in df[C_LOG10_NUM_PERMUTATION].astype(float)]
        pivot_df = df.pivot(columns='Reference', index='Target', values=C_LOG10_NUM_PERMUTATION)
        pivot_df = pivot_df.sort_index(ascending=False)
        label_df = df.pivot(columns='Reference', index='Target', values=C_LABEL)
        label_df = label_df.sort_index(ascending=False)
        # Plot
        ax = sns.heatmap(pivot_df, annot=label_df.values, fmt="", cmap='Reds', vmin=0, vmax=15,
                          annot_kws={'size': font_size}, ax=ax, cbar=is_cbar,
                          cbar_kws={'label': 'log10 number of permutations'})
        ax.figure.axes[-1].yaxis.label.set_size(font_size)
        cbar_ticklabels = ax.figure.axes[-1].get_yticklabels()
        ax.figure.axes[-1].set_yticklabels(cbar_ticklabels, size=font_size)
        ax.tick_params(axis='both', which='major', labelsize=font_size)
        ax.set_xlabel('reference size', size=font_size)
        ax.set_ylabel('target size', size=font_size)
        if title is not None:
            ax.set_title(title, size=font_size + TITLE_FONT_SIZE_INCREMENT)
        else:
            ax.set_title(f'percentile: {percentile}th (10th, 90th)',
                  size=font_size + TITLE_FONT_SIZE_INCREMENT)
        if is_plot:
            plt.show()
        #
        return ax
    
    @staticmethod
    def calculateNoconstraintLog10NumAssignment(reference_size:int, target_size:int)->float:
        """
        Calculate the log10 of the number of assignments (mappings) when no constraints are applied.

        Args:
            reference_size (int): _description_
            target_size (int): _description_

        Returns:
            float: _description_
        """
        count = math.comb(target_size, reference_size)*scipy.special.factorial(reference_size)
        return np.log10(count)
    
    def compareConstraints(self, is_subnet:bool=True)->EvaluateConstraintsResult:
        """
        Selects random reference networks and random targets with the reference network embedded.

        Args:
            is_subnet (bool, optional): reference is a subnet of target if True. Defaults to True.

        Returns:
            EvaluateConstraintsResult
                is_species: True if species constraint
                reference_size: size of the reference network
                target_size: size of the target network
                reaction_df: dataframe of the results
                species_df: dataframe of the results
                short_to_long_dct: mapping of short names to long names
        """
        # Initialize result dictionaries
        species_result_dct:dict = {o.collection_short_name: []
              for o in SpeciesConstraintOptionCollection().iterator()}
        reaction_result_dct:dict = {o.collection_short_name: []
              for o in ReactionConstraintOptionCollection().iterator()}
        # Calculate the number of assignments (mappings) if no constraints are applied
        # Handle no constraint
        reference_size = self.num_species
        target_size = reference_size + self.fill_size
        none_log10_count = self.calculateNoconstraintLog10NumAssignment(reference_size, target_size)
        # Run the simulation
        for _ in range(self.num_iteration):
            if is_subnet:
                num_target_reaction = self.num_reaction + self.fill_size
                num_target_species = self.num_species + self.fill_size
                result = Network.makeRandomReferenceAndTarget(
                      num_reference_reaction=self.num_reaction, num_reference_species=self.num_species,
                      num_target_reaction=num_target_reaction, num_target_species=num_target_species)
                target_network = result.target_network
                reference_network = result.reference_network
            else:
                reference_network = Network.makeRandomNetworkByReactionType(self.num_reaction,
                    self.num_species)
                target_network = Network.makeRandomNetworkByReactionType(target_size, target_size)
            for is_species in [True, False]:
                # Initialize for species or reaction
                if is_species:
                    constraint_cls = SpeciesConstraint
                    default_option_collection = SpeciesConstraintOptionCollection()
                    kwarg = 'species_constraint_option_collection'
                    result_dct = species_result_dct
                else:
                    constraint_cls = ReactionConstraint
                    default_option_collection = ReactionConstraintOptionCollection()
                    kwarg = 'reaction_constraint_option_collection'
                    result_dct = reaction_result_dct
                # Iterate on the options
                for option_collection in default_option_collection.iterator():
                    if option_collection.collection_short_name == cn.NONE:
                        result_dct[option_collection.collection_short_name].append(none_log10_count)
                        continue
                    # Evaluate the constraint on the networks provided
                    reference_constraint = constraint_cls(
                          reference_network.reactant_nmat, reference_network.product_nmat,
                          **{kwarg: option_collection})
                    target_constraint = constraint_cls(
                          target_network.reactant_nmat, target_network.product_nmat,
                          **{kwarg: option_collection})
                    compatibility_collection = reference_constraint.makeCompatibilityCollection(
                          target_constraint).compatibility_collection
                    result_dct[option_collection.collection_short_name].append(
                          compatibility_collection.log10_num_assignment)
        # Construct the dataframe
        species_dimension_result = DimensionResult(dataframe=pd.DataFrame(species_result_dct),
              short_to_long_dct=SpeciesConstraintOptionCollection().short_to_long_dct)
        reaction_dimension_result = DimensionResult(dataframe=pd.DataFrame(reaction_result_dct),
                short_to_long_dct=ReactionConstraintOptionCollection().short_to_long_dct)
        return EvaluateConstraintsResult(
              reference_size=reference_size, 
              target_size=target_size,
              species_dimension_result=species_dimension_result,
              reaction_dimension_result=reaction_dimension_result)
    
    def plotCompareConstraints(self, axs=None, font_size:int=12, is_plot:bool=True,
              **kwargs)->EvaluateConstraintsResult:
        """Bar plot the results of the comparison of constraints.

        Args:
            axs (list, optional): pair of axes. Defaults to None.
            is_plot (bool, optional): plot the results. Defaults to True.
        """
        DCT = {
            'none': 'none',
            'a': 'RC1',
            'b': 'RC2-RC3',
            'a+b': 'RC1-RC3',
            'y': 'SC1-SC2',
            'z': 'SC3-SC4',
            'y+z': 'SC1-SC4',
        }
        if axs is None:
            _, axs = plt.subplots(1, 2, figsize=(10, 5))
        result = self.compareConstraints(**kwargs)
        # Construct the plot
        titles = ['Reactions', 'Species']
        for idx, dimension in enumerate(
              [result.species_dimension_result, result.reaction_dimension_result]):
            ax = axs[idx]
            df = dimension.dataframe
            mean_ser = df.mean(axis=0)
            std_ser = df.std(axis=0)
            mean_ser.plot(kind='bar', ax=ax, legend=False, yerr=std_ser)    
            ax.set_title(titles[idx], size=font_size + TITLE_FONT_SIZE_INCREMENT)
            yticklabels = ax.get_yticklabels()
            ax.set_yticklabels(yticklabels, size=font_size)
            if idx == 0:
                ax.set_ylabel('log10 number of mappings', size=font_size)
            ax.set_xlabel('constraints', size=font_size, labelpad=2)
            xticklabels = [DCT[v] for v in list(df.columns)]
            ax.set_xticklabels(xticklabels, size=font_size, rotation=0)
            # Legends
            if False:
                label_dct = {k: v[3:].replace("_", " ") for k, v in dimension.short_to_long_dct.items()}
                label_dct[cn.NONE] = 'None'
                label_dct['+'.join(dimension.short_to_long_dct.keys())] = 'All'
                labels = [f"{c}: {label_dct[c]}" for c in df.columns]
                pos_arr = [.90, 0.9]
                for label in labels[1:-1]:
                    new_label = label.replace("make ", "")
                    new_label = new_label.replace(" matrix", "")
                    new_label = new_label.replace("n step", "2 step")
                    xpos = pos_arr[0] - 1.01*len(new_label)/100
                    ax.text(xpos, pos_arr[1], new_label, transform=ax.transAxes, fontsize=font_size,
                        ha='center')
                    pos_arr[1] -= 0.05
                    ticklabels = ax.get_xticklabels()
                    ax.set_xticklabels(ticklabels, fontsize=font_size, rotation=0)
        if is_plot:
            plt.show()
        return result
    
    @classmethod
    def calculateOccurrence(cls, species_reaction_sizes:List[Tuple[int, int]],
          identity:str=cn.ID_WEAK, max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT,
          num_replication:int=100, num_iteration:int=1000, is_report:bool=True)->pd.DataFrame:
        """Calculates the probability of occurrence of a reference network in a target network.

        Args:
            species_reaction_sizes (List[Tuple[int, int]]): list of tuples of species and reaction sizes
            max_num_assignment (int, optional): _description_. Defaults to cn.MAX_NUM_ASSIGNMENT.
            num_replication: (int): Number of reference networks evaluated for a single combination
                                    of num_reaction, num_species
            num_iteration: (int): Number of target networks compared with a single reference network

        Returns:
            pd.DataFrame: columns
                num_species: number of species
                num_reaction: number of reactions
                mean probability: probability of occurrence            
                std probability: probability of occurrence
                mean_truncated: probability of truncated occurrence
                poc: -log10(probability of occurrence) rounded to a single digit
        """
        COLUMNS = [cn.D_NUM_SPECIES, cn.D_NUM_REACTION, cn.D_MEAN_PROBABILITY, cn.D_STD_PROBABILITY,
                   cn.D_MEAN_TRUNCATED, POC_SIGNIFICANCE, cn.D_NUM_ITERATION]
        MIN_SIGNIFICANCE = 1e-5
        result_dct:dict = {c: [] for c in COLUMNS}
        for num_species, num_reaction in tqdm.tqdm(species_reaction_sizes,
              desc="pairs", disable=not is_report):
            frac_induceds:list = []
            frac_truncateds:list = []
            for _ in range(num_replication):
                reference_network = Network.makeRandomNetworkByReactionType(num_reaction, num_species)
                calculator = SignificanceCalculatorCore(num_species, num_reaction, num_iteration)
                result = calculator.calculateEqual(reference_network, identity=identity,
                        max_num_assignment=max_num_assignment, is_report=False)
                frac_induceds.append(result.frac_induced)
                frac_truncateds.append(result.frac_truncated)
            result_dct[cn.D_NUM_ITERATION].append(result.num_target_network)
            result_dct[cn.D_NUM_SPECIES].append(num_species)
            result_dct[cn.D_NUM_REACTION].append(num_reaction)
            result_dct[cn.D_MEAN_PROBABILITY].append(np.mean(frac_induceds))
            result_dct[cn.D_STD_PROBABILITY].append(np.std(frac_induceds))
            result_dct[cn.D_MEAN_TRUNCATED].append(np.mean(frac_truncateds))
            mean_induced = max(np.nanmean(frac_induceds), MIN_SIGNIFICANCE)
            if not np.isnan(mean_induced):
                poc = round(-np.log10(mean_induced))
            else:
                poc = np.nan
            result_dct[POC_SIGNIFICANCE].append(poc)
        df = pd.DataFrame(result_dct)
        return df

    @classmethod
    def plotSpeciesReactionHeatmap(cls, 
            dataframe:pd.DataFrame, value_column:str, cbar_title:str="",
            plot_title:str="", ax=None, font_size:int=8, is_cbar:bool=True, vmax:Optional[float]=None,
            is_plot:bool=True)->plt.Axes:
        """Plots a heatmap with the horizontal axis as the number of species and the vertical axis as
            the number of reactions.

        Args:
            dataframe (pd.DataFrame): dataframe with columns num_species, num_reaction, value_column
            cbar_title (str): title of the color bar
            value_column (str): column with the values
            font_size (int, optional): font size of the color bar. Defaults to 8.
            is_cbar (bool, optional): show the color bar. Defaults to True.

        Returns:
            matplotlib.axes._axes.Axes: _description_
        """
        # Construct the plot dataframe
        pivot_df = dataframe.pivot(columns=cn.D_NUM_SPECIES, index=cn.D_NUM_REACTION,
              values=POC_SIGNIFICANCE)
        pivot_df = pivot_df.sort_index(ascending=False)
        # Plot
        if vmax is None:
            vmax = pivot_df.max().max()
        ax = sns.heatmap(pivot_df, annot=True, cmap='Reds', vmin=0, vmax=vmax,
                          annot_kws={'size': font_size}, ax=ax, cbar=is_cbar,
                          cbar_kws={'label': cbar_title})
        ax.figure.axes[-1].yaxis.label.set_size(font_size)
        cbar_ticklabels = ax.figure.axes[-1].get_yticklabels()
        ax.figure.axes[-1].set_yticklabels(cbar_ticklabels, size=font_size)
        ax.tick_params(axis='both', which='major', labelsize=font_size)
        ax.set_xlabel('number of species', size=font_size)
        ax.set_ylabel('number of reactions', size=font_size)
        ax.set_title(plot_title, size=font_size + TITLE_FONT_SIZE_INCREMENT)
        if is_plot:
            plt.show()
        #
        return ax
    

if __name__ == '__main__':
    size = 6
    #ConstraintBenchmark.plotConstraintStudy(size, 9*size, num_iteration=50, is_plot=True)
    benchmark = Benchmark(6, 6, 20)
    _ = benchmark.plotHeatmap(list(range(4, 22, 2)), list(range(10, 105, 5)), percentile=95,
          is_contains_reference=True, num_iteration=500)


'''Representation of a two dimensional array.'''

import numpy as np
import itertools
from typing import Optional

class Matrix(object):

    def __init__(self, array: np.ndarray):
        """
        Args:
            arr (np.array): Stoichiometry matrix. Rows are reactions; columns are species.
        Instance variables:
            num_mat (int): Number of matrices.
            num_row (int): Number of rows.
            num_column (int): Number of columns.
        """
        self.values = array
        self.shape = np.shape(array)
        if len(self.shape) == 2:
            self.num_mat = 1
            self.num_row, self.num_column = np.shape(array)
        elif len(self.shape) == 1:
            self.num_mat = 1
            self.num_row, self.num_column = len(array), 1
        elif len(self.shape) == 3:
            self.num_mat, self.num_row, self.num_column = np.shape(array)
        else:
            self.num_mat, self.num_row, self.num_column = 0, 0, 0
        self._hash:Optional[int] = None  # Deferred execution to improve efficiency

    def __len__(self)->int:
        return self.num_row

    @property
    def hash(self)->int:
        # This is an order dependent hash
        if self._hash is None:
            self._hash = hash(str(self.values))
        return self._hash

    def __repr__(self)->str:
        return str(self.values)

    def __eq__(self, other)->bool:
        if not np.all(np.shape(self.values) == np.shape(other.values)):
            return False
        return np.all(self.values == other.values)  # type: ignore
    
    def isCompatible(self, other)->bool:
        """
        Check if the matrix is compatible with another matrix.
        Args:
            other (Matrix): Another matrix.
        Returns:
            bool: True if the matrix is compatible; False otherwise.
        """
        return bool(np.all(np.shape(self.values) == np.shape(other.values)))
    
    def copy(self)->'Matrix':
        """
        Create a copy of the Matrix.

        Returns:
            Matrix: A copy of the Matrix.
        """
        return Matrix(self.values.copy())
    
    def isPermutablyIdentical(self, other):
        """
        Check if the matrix is permutable with another matrix.
        Args:
            matrix (np.array): Stoichiometry matrix. Rows are reactions; columns are species.
        Returns:
            bool: True if the matrix is permutable; False otherwise.
        """
        row_perm = itertools.permutations(range(self.num_row))
        col_perm = itertools.permutations(range(self.num_column))
        for r in row_perm:
            for c in col_perm:
                if np.all(self.values[r][:, c] == other.matrix):
                    return True
        return False
    
    @classmethod
    def makeTrinaryMatrix(cls, num_row: int=3, num_column: int=2, prob0=1.0/3)->np.ndarray:
        """
        Make a trinary matrix with 0, 1, and -1. No row or column can have all zeros.
        Args:
            nrow (int): Number of rows.
            ncol (int): Number of columns.
        Returns:
            np.array: A trinary matrix.
        """
        prob_other = (1.0 - prob0)/2
        arr = [0, 1, -1]
        prob_arr = [prob0, prob_other, prob_other]
        for _ in range(100):
            matrix = np.random.choice(arr, size=(num_row, num_column), p=prob_arr)
            matrix_sq = matrix*matrix
            is_nozerorow = np.all(matrix_sq.sum(axis=1) > 0)
            is_nozerocol = np.all(matrix_sq.sum(axis=0) > 0)
            if is_nozerorow and is_nozerocol:
                return matrix
        raise RuntimeError('Cannot generate a trinary matrix.')

    def randomize(self):  # type: ignore
        """
        Randomly permutes the rows and columns of the matrix.

        Returns:
            FixedMatrix
        """
        array = self.values.copy()
        row_perm = np.random.permutation(self.num_row)
        col_perm = np.random.permutation(self.num_column)
        return Matrix(array[row_perm][:, col_perm])  # type: ignore


import pySubnetSB.constants as cn  # type: ignore
from pySubnetSB.network_base import NetworkBase # type: ignore
from pySubnetSB.network import Network # type: ignore
from pySubnetSB.named_matrix import NamedMatrix # type: ignore
from pySubnetSB import util # type: ignore
from pySubnetSB.assignment_pair import AssignmentPair # type: ignore

import os
from pynauty import Graph  # type: ignore
import matplotlib.pyplot as plt
import numpy as np
import copy
import tellurium as te  # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
NETWORK_NAME = "test"
BIG_NETWORK = """
J0: S1 -> S1 + S2; k1*S1;
J1: S2 -> S3; k2*S2;
J2: S3 -> ; k3*S3;
J3: S3 -> S4; k4*S3;
J4: S4 -> S5; k5*S4;

k1 = 1
k2 = 1
k3 = 1
k4 = 1
k5 = 1
S1 = 10
S2 = 0
S3 = 0
S4 = 0
S5 = 0
"""
NETWORK1 = """
J0: S1 -> S1 + S2; k1*S1;
J1: S2 -> S3; k2*S2;

k1 = 1
k2 = 1
S1 = 10
S2 = 0
S3 = 0
"""
NETWORK2 = """
// Strong structural identity with Network1
Ja: S2 -> S2 + S1; k1*S2;
Jb: S1 -> S3; k2*S1;

k1 = 3
k2 = 1
S1 = 1
S2 = 0
S3 = 0
"""
NETWORK3 = """
// Weak sructural identity with Network1
Ja: 2 S1 -> 2 S1 + S2; k1*S1*S1
Jb: S2 -> S3; k2*S1;

k1 = 3
k2 = 1
S1 = 1
S2 = 0
S3 = 0
"""
NETWORK4 = """
// Not structurally identical with Network1
Ja: S1 -> S2; k1*S2;
Jb: S1 -> S3; k2*S1;

k1 = 3
k2 = 1
S1 = 1
S2 = 0
S3 = 0
"""
NETWORK = NetworkBase.makeFromAntimonyStr(NETWORK1, network_name=NETWORK_NAME)


class TestNetwork(unittest.TestCase):

    def setUp(self):
        self.network = copy.deepcopy(NETWORK)

    @util.timeit
    def testConstrutor(self):
        if IGNORE_TEST:
            return
        self.assertEqual(self.network.network_name, NETWORK_NAME)
        for mat in [self.network.reactant_nmat, self.network.product_nmat]:
            self.assertTrue(isinstance(mat, NamedMatrix))
        self.assertTrue("int" in str(type(self.network.network_hash)))
        self.assertTrue("int" in str(type(self.network.network_hash)))
    
    @util.timeit
    def testCopyEqual(self):
        if IGNORE_TEST:
            return
        network = self.network.copy()
        self.assertEqual(self.network, network)

    @util.timeit
    def testIsMatrixEqual(self):
        if IGNORE_TEST:
            return
        network = NetworkBase.makeFromAntimonyStr(NETWORK1, network_name=NETWORK_NAME)
        network3 = NetworkBase.makeFromAntimonyStr(NETWORK3, network_name=NETWORK_NAME)
        self.assertTrue(network.isMatrixEqual(network3, identity=cn.ID_WEAK))
        self.assertFalse(network.isMatrixEqual(network3, identity=cn.ID_STRONG))
        self.assertTrue(network.isMatrixEqual(network, identity=cn.ID_WEAK))
        self.assertTrue(network.isMatrixEqual(network, identity=cn.ID_STRONG))
        
    @util.timeit
    def testRandomlyPermuteTrue(self):
        if IGNORE_TEST:
            return
        def test(size, num_iteration=500):
            reactant_arr = np.random.randint(0, 3, (size, size))
            product_arr = np.random.randint(0, 3, (size, size))
            network = NetworkBase(reactant_arr, product_arr)
            for _ in range(num_iteration):
                new_network, assignment_pair = network.permute()
                if network == new_network:
                    continue
                original_network, _ = new_network.permute(assignment_pair=assignment_pair)
                self.assertTrue(network.isEquivalent(original_network))
                self.assertEqual(network.num_species, new_network.num_species)
                self.assertEqual(network.num_reaction, new_network.num_reaction)
                self.assertEqual(network.network_hash, new_network.network_hash)
                self.assertEqual(network.network_hash, new_network.network_hash)
        #
        test(3)
        test(30)

    @util.timeit
    def testRandomlyPermuteFalse(self):
        if IGNORE_TEST:
            return
        # Randomly change a value in the reactant matrix
        def test(size, num_iteration=500):
            network_hashes = []
            for _ in range(num_iteration):
                network = NetworkBase.makeRandomNetworkByReactionType(num_reaction=3*size, num_species=size)
                network_hashes.append(network.network_hash)
            num_distinct_weak_hashes = len(set(network_hashes))
            frac_distinct_hashes =  num_distinct_weak_hashes/ num_iteration
            #print(f"distinct hashes: {frac_distinct_hashes}")
            self.assertGreater(frac_distinct_hashes, 0.9)
        #
        test(4)

    @util.timeit
    def testIsStructurallyCompatible(self):
        if IGNORE_TEST:
            return
        network1 = NetworkBase.makeFromAntimonyStr(NETWORK1, network_name=NETWORK_NAME)
        network2 = NetworkBase.makeFromAntimonyStr(NETWORK2, network_name=NETWORK_NAME)
        network3 = NetworkBase.makeFromAntimonyStr(NETWORK3, network_name=NETWORK_NAME)
        network4 = NetworkBase.makeFromAntimonyStr(NETWORK4, network_name=NETWORK_NAME)
        self.assertFalse(network1.isStructurallyCompatible(network4))
        self.assertTrue(network1.isStructurallyCompatible(network2))
        self.assertTrue(network1.isStructurallyCompatible(network3, identity=cn.ID_WEAK))

    @util.timeit
    def testPrettyPrintReaction(self):
        if IGNORE_TEST:
            return
        network = NetworkBase.makeFromAntimonyStr(NETWORK3, network_name="Network3")
        stg = network.prettyPrintReaction(0)
        self.assertTrue("2.0 S1 -> 2.0 S1 + S2" in stg)

    @util.timeit
    def testMakeRandomNetworkFromReactionType(self):
        if IGNORE_TEST:
            return
        for _ in range(100):
            size = np.random.randint(3, 20)
            network = NetworkBase.makeRandomNetworkByReactionType(size, is_exact=True)
            self.assertEqual(network.num_species, size)
            eval_arr = np.vstack([network.reactant_nmat.values, network.product_nmat.values])
            # Verify that all reactions have at least one participant
            sum_arr = np.sum(eval_arr, axis=0)
            self.assertTrue(np.all(sum_arr > 0))

    @util.timeit
    def testMakeRandomNetworkFromReactionType2(self):
        if IGNORE_TEST:
            return
        size = 3
        species_names = ["S1", "S5", "S10"]
        reaction_names = ['JJ1', 'JJ2', 'JJ3']
        network = NetworkBase.makeRandomNetworkByReactionType(size, species_names=species_names,
              reaction_names=reaction_names, is_exact=True)
        self.assertEqual(len(species_names), network.num_species)
        for species_name in species_names:
            self.assertTrue(species_name in network.species_names)
        for reaction_name in reaction_names:
            self.assertTrue(reaction_name in network.reaction_names)
    
    @util.timeit
    def testMakeRandomReferenceAndTarget(self):
        if IGNORE_TEST:
            return
        size = 4
        for _ in range(10):
            result = Network.makeRandomReferenceAndTarget(size, 2*size)
            reference_network = result.reference_network
            target_network = result.target_network
            self.assertEqual(reference_network.num_species, size)
            self.assertEqual(target_network.num_species, 2*size)
            self.assertEqual(reference_network.num_reaction, size)
            self.assertEqual(target_network.num_reaction, 2*size)
            if not reference_network.isStructurallyIdentical(target_network, is_subnet=True,
                  is_report=True, num_process=1, max_num_assignment=int(1e18), is_all_valid_assignment=False):
                import pdb; pdb.set_trace()
            self.assertTrue(reference_network.isStructurallyIdentical(target_network, is_subnet=True))

    @util.timeit
    def testToFromSeries(self):
        if IGNORE_TEST:
            return
        series = self.network.toSeries()
        serialization_str = self.network.seriesToJson(series)
        network = NetworkBase.deserialize(serialization_str)
        self.assertTrue(self.network == network)

    @util.timeit
    def testMakePynautyNetwork(self):
        if IGNORE_TEST:
            return
        graph = self.network.makePynautyNetwork()
        self.assertTrue(isinstance(graph, Graph))

    @util.timeit
    def testCSVNetwork(self):
        if IGNORE_TEST:
            return
        csv_format = self.network.makeCSVNetwork()
        self.assertTrue(isinstance(csv_format, str))
        num_line = csv_format.count("\n")
        num_sep = csv_format.count(">")
        self.assertEqual(num_line+1,num_sep + self.network.num_species + self.network.num_reaction)
    
    @util.timeit
    def testGetGraphDct(self):
        if IGNORE_TEST:
            return
        num_iteration = 100
        size =10 
        for _ in range(num_iteration):
            for identity in cn.ID_LST:
                network = NetworkBase.makeRandomNetwork(size, size)
                graph_descriptor = network.getGraphDescriptor(identity=identity)
                vertex_dct = graph_descriptor.vertex_dct
                label_dct = graph_descriptor.label_dct
                self.assertTrue(isinstance(vertex_dct, dict))
                self.assertTrue(isinstance(vertex_dct[0], list))
                if identity == cn.ID_STRONG:
                    num_edge = network.reactant_nmat.values.sum() + network.product_nmat.values.sum()
                else:
                    num_edge = np.abs(network.standard_nmat.values).sum()
                num_graph_edge = np.sum([len(e) for e in vertex_dct.values()])
                self.assertEqual(num_edge, num_graph_edge)
                #
                num_species = np.sum([v == 'species' for v in label_dct.values()])
                self.assertEqual(num_species, network.num_species)
                num_reaction = np.sum([v == 'reaction' for v in label_dct.values()])
                self.assertEqual(num_reaction, network.num_reaction)

    @util.timeit
    def testFill(self):
        if IGNORE_TEST:
            return
        for fill_size in range(1, 5):
            network = NetworkBase.makeRandomNetworkByReactionType(5, 5)
            filled_network = network.fill(num_fill_reaction=fill_size, num_fill_species=fill_size)
            self.assertLessEqual(filled_network.num_species, network.num_species + fill_size)
            self.assertEqual(filled_network.num_reaction, network.num_reaction + fill_size)
        #
        with self.assertRaises(ValueError):
            network.fill(num_fill_reaction=0, num_fill_species=0)

    @util.timeit

    def testIsBoundaryNetwork(self):
        if IGNORE_TEST:
            return
        boundary_network = """
            R1: A -> ; k1*A
            R2:  -> B; k2
            k1 = 0.1; k2 = 0.2
            A = 0; B = 0
        """
        network = NetworkBase.makeFromAntimonyStr(boundary_network)
        self.assertTrue(network.isBoundaryNetwork())
        #
        boundary_network = """
            R1: A -> ; k1*A
            R3: A -> C; k1*A
            R2: B -> A; k2*B
            k1 = 0.1; k2 = 0.2
            A = 0; B = 0
        """
        network = NetworkBase.makeFromAntimonyStr(boundary_network)
        self.assertFalse(network.isBoundaryNetwork())

    def testMakeFromAntimonyStrRoadrunner(self):
        if IGNORE_TEST:
            return
        roadrunner = te.loada(BIG_NETWORK)
        network = NetworkBase.makeFromAntimonyStr(None, roadrunner=roadrunner)
        self.assertGreater(network.num_species, 0)

    def testMakeFromSBMLFile(self):
        if IGNORE_TEST:
            return
        PATH = os.path.join(cn.TEST_DIR, "xml_files/BIOMD0000000033.xml")
        network = NetworkBase.makeFromSBMLFile(PATH)
        self.assertGreater(network.num_species, 0)

    def testMakeInducedNetwork(self):
        if IGNORE_TEST:
            return
        species_assignment = np.array(range(self.network.num_species))
        reaction_assignment = np.array(range(self.network.num_reaction))
        assignment_pair = AssignmentPair(species_assignment, reaction_assignment)
        induced_network = self.network.makeInducedNetwork(assignment_pair)
        self.assertTrue(self.network.isEquivalent(induced_network))

    def testMakeMatricesForIdentity(self):
        if IGNORE_TEST:
            return
        network1 = NetworkBase.makeFromAntimonyStr(NETWORK1, network_name=NETWORK_NAME)
        network4 = NetworkBase.makeFromAntimonyStr(NETWORK4, network_name=NETWORK_NAME)
        #####
        def test(network, identity, expected_bool):
            reactant_nmat, product_nmat = network.makeMatricesForIdentity(identity=identity)
            self.assertEqual(reactant_nmat == network.reactant_nmat, expected_bool)
            self.assertEqual(product_nmat == network.product_nmat, expected_bool)
        #####
        test(network1, cn.ID_STRONG, True)
        test(network1, cn.ID_WEAK, False)
        test(network4, cn.ID_STRONG, True)
        test(network4, cn.ID_WEAK, True)


if __name__ == '__main__':
    unittest.main(failfast=False)


'''Constraint species.'''

"""
Implements the calculation of categorical and enumerated constraints for species.
  Enumerated constraints: 
    counts of reaction types in which a species is a reactant or product
    number of autocatalysis reactions for a species.
"""

from pySubnetSB.named_matrix import NamedMatrix # type: ignore
from pySubnetSB.constraint import Constraint, ReactionClassification, NULL_NMAT # type: ignore
import pySubnetSB.util as util  # type: ignore
from pySubnetSB.constraint_option_collection import SpeciesConstraintOptionCollection # type: ignore

import numpy as np
from typing import List



#####################################
class SpeciesConstraint(Constraint):

    def __init__(self, reactant_nmat:NamedMatrix, product_nmat:NamedMatrix, is_subnet:bool=True,
              species_constraint_option_collection:SpeciesConstraintOptionCollection=SpeciesConstraintOptionCollection()):  
        """
        Args:
            reactant_nmat (NamedMatrix)
            product_nmat (NamedMatrix)
            is_subnet (bool, optional) Consider self as a subset of other.
        """
        super().__init__(reactant_nmat=reactant_nmat, product_nmat=product_nmat, is_subnet=is_subnet)
        #
        self._is_initialized = False
        self._numerical_enumerated_nmat = NULL_NMAT
        self._categorical_nmat = NULL_NMAT
        self._bitwise_enumerated_nmat = NULL_NMAT
        self._one_step_nmat = NULL_NMAT
        self.species_constraint_options = species_constraint_option_collection

    ################# OVERLOADED PARENT CLASS METHODS #################
    @property
    def row_names(self)->List[str]:
        return self.reactant_nmat.row_names
    
    @property
    def description(self)->str:
        return "species"

    @property
    def numerical_enumerated_nmat(self)->NamedMatrix:
        if not self._is_initialized:
            matrices = []
            if self.species_constraint_options.isTrue("is_reactant_product_count_constraint_matrix"):
                matrices.append(self._makeReactantProductCountConstraintMatrix())
            if self.species_constraint_options.isTrue("is_autocatalysis_constraint"):
                matrices.append(self._makeAutocatalysisConstraint())
            if self.species_constraint_options.isTrue("is_reactant_product_constraint_matrix"):
                matrices.append(self._makeReactantProductConstraintMatrix())
            if self.species_constraint_options.isTrue("is_successor_predecessor_constraint_matrix"):
                matrices.append(self.makeSuccessorPredecessorConstraintMatrix())
            if self.species_constraint_options.isTrue("is_n_step_constraint_matrix"):
                matrices.append(self.makeNStepConstraintMatrix(num_step=2))
            #
            if len(matrices) == 0:
                self._numerical_enumerated_nmat = NULL_NMAT
            else:
                self._numerical_enumerated_nmat = NamedMatrix.hstack(matrices)
            self._is_initialized = True
        return self._numerical_enumerated_nmat

    @property
    def categorical_nmat(self)->NamedMatrix:
        if self._categorical_nmat is NULL_NMAT:
              #self._categorical_mat = self._makeAutocatalysisConstraint()  # Categorical version
              pass
        return self._categorical_nmat

    @property
    def bitwise_enumerated_nmat(self)->NamedMatrix:
        if self._bitwise_enumerated_nmat is NULL_NMAT:
            #self._bitwise_enumerated_nmat = self._makeBitwiseReactantProductConstraintMatrix()
            pass
        return self._bitwise_enumerated_nmat
    
    @property
    def one_step_nmat(self)->NamedMatrix:
        """Calculates the successor matrix for the species monopartite graph.

        Returns:
            np.ndarray: _description_
        """
        # Create the monopartite graph
        if self._one_step_nmat is NULL_NMAT:
            incoming_arr = self.reactant_nmat.values
            outgoing_arr = self.product_nmat.values
            arr = np.sign(np.matmul(incoming_arr, outgoing_arr.T)).astype(int)
            self._one_step_nmat = NamedMatrix(arr, row_names=self.reactant_nmat.row_names,
                                row_description='species',
                                column_description='species',
                                column_names=self.reactant_nmat.row_names)
        return self._one_step_nmat
    
    @property
    def to_reaction_arr(self)->np.ndarray:
        return self.reactant_nmat.values
    
    ##################################

    def __repr__(self)->str:
        return "Species--" + super().__repr__()
    
    def _makeReactantProductConstraintMatrix(self)->NamedMatrix:
        """Make constraints for the reaction classification. These are {R, P} X {ReactionClassifications}
        where R is reactant and P is product.

        Returns:
            NamedMatrix: Rows are species, columns are constraints
        """
        reaction_classifications = [str(c) for c in ReactionClassification.getReactionClassifications()]
        reactant_arrays = []
        product_arrays = []
        for i_species in range(self.num_species):
            reactant_array = np.zeros(len(reaction_classifications))
            product_array = np.zeros(len(reaction_classifications))
            for i_reaction in range(self.num_reaction):
                i_constraint_str = str(self.reaction_classification_arr[i_reaction])
                idx = reaction_classifications.index(i_constraint_str)
                if self.reactant_nmat.values[i_species, i_reaction] > 0:
                    reactant_array[idx] += 1
                if self.product_nmat.values[i_species, i_reaction] > 0:
                    product_array[idx] += 1
            reactant_arrays.append(reactant_array)
            product_arrays.append(product_array)
        # Construct full array and labels
        arrays = np.concatenate([reactant_arrays, product_arrays], axis=1)
        reactant_labels = [f"r_{c}" for c in reaction_classifications]
        product_labels = [f"p_{c}" for c in reaction_classifications]
        column_labels = reactant_labels + product_labels
        # Make the NamedMatrix
        named_matrix = NamedMatrix(np.array(arrays), row_names=self.reactant_nmat.row_names,
                           row_description='species',
                           column_description='constraints',
                           column_names=column_labels)
        return named_matrix
    
    def _makeReactantProductCountConstraintMatrix(self)->NamedMatrix:
        """Calculates the count of reactions in which a species is a reactant or product.

        Returns:
            NamedMatrix: Rows are species, columns are constraints
        """
        reactant_count_arr = np.sum(self.reactant_nmat.values, axis=1)
        product_count_arr = np.sum(self.product_nmat.values, axis=1)
        array = np.array([reactant_count_arr, product_count_arr]).T
        column_labels = ["reactant_count", "product_count"]
        # Make the NamedMatrix
        named_matrix = NamedMatrix(array, row_names=self.reactant_nmat.row_names,
                           row_description='species',
                           column_description='constraints',
                           column_names=column_labels)
        return named_matrix
    
    def _makeBitwiseReactantProductConstraintMatrix(self)->NamedMatrix:
        """For each reach, construct two bit vectors, one for reactants and one for products,
        that species the type of reactions in which the species is present (as a reactant or product).

        Returns:
            NamedMatrix: Rows are species, columns are bit vectors for reactions where species is a reactant or product.
        """
        reactant_bit_vectors:List[int] = []
        product_bit_vectors:List[int] = []
        for i_species in range(self.num_species):
            reactant_reaction_type_bit = 0
            product_reaction_type_bit = 0
            for i_reaction in range(self.num_reaction):
                index = self.reaction_classification_arr[i_reaction].index
                if self.reactant_nmat.values[i_species, i_reaction] > 0:
                    reactant_reaction_type_bit |= 1 << index
                if self.product_nmat.values[i_species, i_reaction] > 0:
                    product_reaction_type_bit |= 1 << index
            reactant_bit_vectors.append(reactant_reaction_type_bit)
            product_bit_vectors.append(product_reaction_type_bit)
        # Construct full array and labels
        array = np.array([reactant_bit_vectors, product_bit_vectors]).T
        column_labels = ["reactant_bit_vector", "product_bit_vector"]
        # Make the NamedMatrix
        named_matrix = NamedMatrix(array, row_names=self.reactant_nmat.row_names,
                           row_description='species',
                           column_description='constraints',
                           column_names=column_labels)
        return named_matrix
    
    def _makeAutocatalysisConstraint(self)->NamedMatrix:
        """Indicates if the species is involved in an autocatalysis reaction.

        Returns:
            NamedMatrix. is_autocatalysis
        """
        column_names =  ['is_autocatalysis']
        array = self.reactant_nmat.values * self.product_nmat.values > 0
        vector = np.sum(array, axis=1)
        vector = np.reshape(vector, (len(vector), 1)).astype(int)
        #vector = np.sign(vector)
        named_matrix = NamedMatrix(vector, row_names=self.reactant_nmat.row_names,
                           row_description='species',
                           column_description='constraints',
                           column_names=column_names)
        return named_matrix


'''Evaluates collisions in the identity hash.'''

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.network import Network      # type: ignore

import collections
import matplotlib.pyplot as plt # type: ignore
import numpy as np
import pandas as pd  # type: ignore
import matplotlib.pyplot as plt
import seaborn as sns  # type: ignore
from typing import List, Optional

NUM_NETWORK = 'num_network'
NUM_COLLISION = 'num_collision'
NUM_UNIQUE_HASH = 'num_unique_hash'
MAX_NETWORK_IN_HASH = 'max_network_in_hash'
NUM_SPECIES = 'num_species'
NUM_REACTION = 'num_reaction'
PERCENT_COLLISION = 'percent_collision'
#  num_network: number of networks simulated
#  num_collision: number of collisions in total
#  num_unique_hash: number of distinct hash values
#  max_network_in_hash: maximum number of networks in a hash
#  num_species: number of species
#  num_reaction: number of reactions
#  frac_collision: fraction of collisions
COLUMNS = [NUM_SPECIES, NUM_REACTION, NUM_NETWORK, NUM_COLLISION, NUM_UNIQUE_HASH, MAX_NETWORK_IN_HASH,
      PERCENT_COLLISION]
TITLE_FONT_SIZE_INCREMENT = 2
NAME_DCT = {NUM_SPECIES: 'number of species', NUM_REACTION: 'number of reactions',
            NUM_COLLISION: 'number collision',
      NUM_UNIQUE_HASH: 'number unique hash', MAX_NETWORK_IN_HASH: 'max network in hash',
      PERCENT_COLLISION: 'percent collision'}


class IdentityHashBenchmark(object):
    def __init__(self, species_nums:List[int], reaction_nums:List[int], num_network:int=1000):
        """
        Args:
            species_nums: number of species
            reaction_nums: number of reactions
            num_iteration: number of networks simulated for a combination of species and reactions
        """
        self.species_nums = species_nums
        self.reaction_nums = reaction_nums
        self.num_network = num_network

    def calculateHashStatistics(self)->pd.DataFrame:
        """
        Calculates the number of collisions in the identity hash.

        Randomly generates networks and calculates the number of collisions in the identity hash.
        Eliminates strongly identical networks from collisions.

        Returns:
            pd.DataFrame: dataframe
                num_species: number of species
                num_reaction: number of reactions
                num_network: number of networks simulated
                num_collision: number of collisions in total
                num_unique_hash: number of distinct hash values
                max_network_in_hash: maximum number of networks in a hash
        """
        # Initialize result dictionaries
        dct:dict = {c: [] for c in COLUMNS}
        for num_species in self.species_nums:
            for num_reaction in self.reaction_nums:
                networks = [Network.makeRandomNetworkByReactionType(
                    num_species=num_species, num_reaction=num_reaction)
                      for _ in range(self.num_network)]
                # Group networks by hash
                hash_dct:dict = {}
                for network in networks:
                    if network.network_hash in hash_dct:
                        hash_dct[network.network_hash].append(network)
                    else:
                        hash_dct[network.network_hash] = [network]
                # Eliminate duplicates
                unduplicated_network_dct:dict = {}
                for key in hash_dct.keys():
                    unduplicated_network_dct[key] = [hash_dct[key][0]]
                    for same_hash_network in hash_dct[key][1:]:
                        result = same_hash_network.isStructurallyIdentical(unduplicated_network_dct[key][0],
                                is_subnet=False, identity=cn.ID_STRONG)
                        if result.is_truncated:
                            print('Truncated')
                        if not result:
                            unduplicated_network_dct[key].append(same_hash_network)
                # Calculate statistics
                count_arr = np.array([len(v) for v in unduplicated_network_dct.values()])
                max_network_in_hash = max(count_arr)
                num_collision = np.sum([c for c in count_arr if c > 1])
                num_unique_hash = len(count_arr)
                # Record the results
                dct[NUM_SPECIES].append(num_species)
                dct[NUM_REACTION].append(num_reaction)
                dct[NUM_NETWORK].append(self.num_network)
                dct[NUM_COLLISION].append(num_collision)
                dct[NUM_UNIQUE_HASH].append(num_unique_hash)
                dct[MAX_NETWORK_IN_HASH].append(max_network_in_hash)
                percent_collision = round(100*num_collision / np.sum(count_arr))
                dct[PERCENT_COLLISION].append(percent_collision)
        # Create the dataframe
        df = pd.DataFrame(dct)
        return df
    
    def plotHashStatistics(self, statistic_name:str=PERCENT_COLLISION, title:str="",
          font_size:int=12, hash_statistics_df:Optional[pd.DataFrame]=None, is_plot=True)->pd.DataFrame:
        """
        Plots the number of collisions in the identity hash.

        Args:
            statistic_name: column name in the dataframe
            title: title of the plot
            font_size: font size
            hash_statistics_df: dataframe for data used
            is_plot: if True, plot the DatFrame
        """
        if hash_statistics_df is None:
            hash_statistics_df = self.calculateHashStatistics()
        # Create a pivot table
        plot_df = hash_statistics_df.pivot_table(index=NUM_SPECIES,
              columns=NUM_REACTION, values=statistic_name)
        plot_df = plot_df.sort_index(ascending=False)
        # Plot the heatmap
        ax = sns.heatmap(plot_df, cmap='Reds', vmin=0, vmax=100,
                          annot_kws={'size': font_size}, annot=True,
                          cbar_kws={'label': NAME_DCT[statistic_name]})
        ax.tick_params(axis='both', which='major', labelsize=font_size)
        ax.set_xlabel(NAME_DCT[NUM_REACTION], size=font_size)
        ax.set_ylabel(NAME_DCT[NUM_SPECIES], size=font_size)
        ax.set_title(title, size=font_size + TITLE_FONT_SIZE_INCREMENT)
        ax.figure.axes[-1].yaxis.label.set_size(font_size)
        cbar_ticklabels = ax.figure.axes[-1].get_yticklabels()
        ax.figure.axes[-1].set_yticklabels(cbar_ticklabels, size=font_size)
        if is_plot:
            plt.show()
        return hash_statistics_df


from pySubnetSB.model_serializer import ModelSerializer  # type: ignore
from pySubnetSB.network_collection import NetworkCollection  # type: ignore
import pySubnetSB.constants as cn   # type: ignore

import os
import unittest


IGNORE_TEST = False
IS_PLOT = False
MODEL_DIRECTORY = "oscillators"
SERIALIZATION_FILE = os.path.join(cn.DATA_DIR, 'oscillators_serializers.txt')


#############################
# Tests
#############################
class TestModelSerializer(unittest.TestCase):

    def setUp(self):
        self.model_serializer = ModelSerializer.makeOscillatorSerializer(MODEL_DIRECTORY,
              parent_directory=cn.TEST_DIR)
        self.remove()

    def tearDown(self):
        self.remove()

    def remove(self):
        if os.path.exists(SERIALIZATION_FILE):
            os.remove(SERIALIZATION_FILE)

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertTrue(self.model_serializer.model_directory.split("/")[-1] == MODEL_DIRECTORY)

    def testSerializeDeserialize(self):
        if IGNORE_TEST:
            return
        self.model_serializer.serialize()
        model_serializer = ModelSerializer.makeOscillatorSerializer(MODEL_DIRECTORY, parent_directory=cn.TEST_DIR)
        network_collection = model_serializer.deserialize()
        self.assertTrue(isinstance(network_collection, NetworkCollection))
        ffiles = [f for f in os.listdir(os.path.join(cn.TEST_DIR, MODEL_DIRECTORY))
              if not f.endswith('txt')]
        self.assertTrue(len(ffiles) - len(network_collection) <= 1)

    def testDeserializeWithNames(self):
        if IGNORE_TEST:
            return
        SIZE = 5
        self.model_serializer.serialize()
        network_collection = self.model_serializer.deserialize()
        names = [n.network_name for n in network_collection.networks[0:SIZE]]
        new_network_collection = self.model_serializer.deserialize(model_names=names)
        self.assertEqual(len(new_network_collection), SIZE)
        for idx, name in enumerate(names):
            self.assertEqual(name, new_network_collection.networks[idx].network_name)

        

if __name__ == '__main__':
    unittest.main()


'''A container for collections of structurally identical networks and their statistics.'''
"""
Has a string representation and can construct from its string representation.
"""


from pySubnetSB import constants as cn # type: ignore
from pySubnetSB.processed_network import ProcessedNetwork  # type: ignore
from pySubnetSB.assignment_pair import AssignmentPair  # type: ignore

import json
import collections
import numpy as np
import pandas as pd  # type: ignore
from typing import List, Optional

Repr = collections.namedtuple('Repr',
     ['identity', 'hash_val', 'processed_networks'])


class ProcessedNetworkCollection(object):
    # Collection of networks that are structurally identical

    def __init__(self, processed_networks:List[ProcessedNetwork],
                 identity:str=cn.ID_WEAK, hash_val:int=-1,
                 antimony_directory:Optional[str]=None):
        self.processed_networks = processed_networks  # type: ignore
        self.identity = identity
        self.hash_val = hash_val
        self.antimony_directory = antimony_directory # Directory where the network is stored

    @property
    def is_indeterminate(self)->bool:
        return any([n.is_indeterminate for n in self.processed_networks])

    @property
    def processing_time(self)->float:
        return np.sum([c.processing_time for c in self.processed_networks])

    def copy(self)->'ProcessedNetworkCollection':
        return ProcessedNetworkCollection([cn.copy() for cn in self.processed_networks],
                identity=self.identity,
                hash_val=self.hash_val)

    def __eq__(self, other:object)->bool:
        if not isinstance(other, ProcessedNetworkCollection):
            return False
        if self.identity != other.identity:
            import pdb; pdb.set_trace()
            return False
        if self.hash_val != other.hash_val:
            import pdb; pdb.set_trace()
            return False
        for net1, net2 in zip(self.processed_networks, other.processed_networks):
            if not net1 == net2:
                import pdb; pdb.set_trace()
                return False
        return True
    
    def isSubset(self, other:object)->bool:
        # Is this a subset of other?
        if not isinstance(other, ProcessedNetworkCollection):
            return False
        for this_network in self.processed_networks:
            is_found = False
            for other_network in other.processed_networks:
                if this_network == other_network:
                    is_found = True
                    break
            if not is_found:
                return False
        return True

    def __len__(self)->int:
        return len(self.processed_networks)
    
    def __repr__(self)->str:
        # Summary of the object
        if self.identity == cn.ID_STRONG:
            prefix = cn.IDENTITY_PREFIX_STRONG
        else:
            prefix = cn.IDENTITY_PREFIX_WEAK
        names = [c.network_name for c in self.processed_networks]
        processed_networks_str = cn.NETWORK_DELIMITER.join(names)
        result = f"{prefix}{self.processing_time}--{self.hash_val}--{processed_networks_str}"
        return result

    def add(self, processed_network:ProcessedNetwork):
        self.processed_networks.append(processed_network)

    def serialize(self)->str:
        """Creates a JSON string for the object.

        Returns:
            str
        """
        dct = {cn.S_ID: str(self.__class__),
               cn.S_PROCESSED_NETWORKS: [c.serialize() for c in self.processed_networks],
               cn.S_IDENTITY: self.identity,
               cn.S_HASH_VAL: int(self.hash_val),  # Cannot serialize numpy.int64
               cn.S_ANTIMONY_DIRECTORY: self.antimony_directory,
        }
        return json.dumps(dct)
    
    @classmethod
    def deserialize(cls, serialization_str)->'ProcessedNetworkCollection':
        """Creates a ProcessedNetworkCollection from a JSON serialization string.

        Args:
            serialization_str

        Returns:
            ProcessedNetworkCollection
        """
        dct = json.loads(serialization_str)
        if not cls.__name__ in dct[cn.S_ID]:
            raise ValueError(f"Expected {cls} but got {dct[cn.S_ID]}")
        identity = dct[cn.S_IDENTITY]
        antimony_directory =  dct[cn.S_ANTIMONY_DIRECTORY]
        processed_networks = [ProcessedNetwork.deserialize(s) for s in dct[cn.S_PROCESSED_NETWORKS]]
        hash_val = dct[cn.S_HASH_VAL]
        processed_network_collection = ProcessedNetworkCollection(processed_networks, identity=identity,
                                          hash_val=hash_val, antimony_directory=antimony_directory)
        return processed_network_collection


'''Worker runs in a single process, reads workunits, calls SubnetFinder, and writes a results checkpoint file.'''

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.subnet_finder import SubnetFinder # type: ignore
from pySubnetSB.model_serializer import ModelSerializer # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.worker_checkpoint_manager import WorkerCheckpointManager  # type: ignore
from pySubnetSB.subnet_finder_workunit_manager import SubnetFinderWorkunitManager # type: ignore

import pandas as pd  # type: ignore
import sys
import time
from typing import List, Optional

REPORT_INTERVAL = 1000


def executeWorker(worker_idx:int, workunit_path, num_worker:int, checkpoint_path:str,
        reference_serialization_path:str, target_serialization_path:str,
        identity:str=cn.ID_STRONG, is_report:bool=True,
        is_initialize:bool=True,
        max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT,
        checkpoint_interval:int=10,
        is_allow_exit:bool=True)->None:
    """
    Processes pairs of reference and target networks. Networks are processed
    based on the index provided in the queue. A worker creates checkpoint file that is a
    serialization of the DataFrame structured as described in SubnetFinder
    (cn.FINDER_DATAFRAME_COLUMNS).
    The checkpoint file is used to recover the state of the worker, including its completion.

    Args:
        worker_idx (int): Index of the worker
        workunit_path (str): Path to the workunit file
        num_worker (int): Total number of workers
        checkpoint_path (str): Output path for the consoldated checkpoint files
        reference_serialization_path (str): Path to the serialized reference networks
        target_serialization_path (str): Path to the serialized target networks
        identity (str): Identity type
        is_report (float): Worker 0 reports progress
        is_initialize (bool): If True, initialize the checkpoint
        max_num_assignment (int): Maximum number of assignment pairs
        checkpoint_interval (int): Interval for checkpointing
        is_allow_exit (bool): If True, exit after completion (faster handling of queue merge for parallel)
    """
    LAST_TIME = "last_time"
    CHECKPOINT_TIME = 5*60  # 5 minutes
    global_dct = {LAST_TIME: time.time()}
    #####
    def printPerformanceInformation(msg:str):
        # Print performance information
        if False:
            last_time = global_dct[LAST_TIME]
            now_time = time.time()
            duration = now_time - last_time
            new_msg = f"**{msg}: {duration} seconds."
            print(new_msg, flush=True)
            global_dct[LAST_TIME] = now_time
    #####
    def getNetworks(serialization_path:str)->List[Network]:
        """
        Obtains the networks from a serialization file.

        Args:
            serialization_path (str): Path to the serialization file

        Returns:
            List[Network]: List of networks
        """
        serializer = ModelSerializer(None, serialization_path)
        collection = serializer.deserialize()
        return collection.networks
    #####
    is_report_worker = is_report if worker_idx == 0 else False
    # Set up the checkpoint manager for this worker
    worker_checkpoint_path = WorkerCheckpointManager.makeWorkerCheckpointPath(checkpoint_path,
          worker_idx)
    worker_checkpoint_manager = WorkerCheckpointManager(worker_checkpoint_path,
          is_report=is_report_worker, is_initialize=is_initialize)
    full_worker_df = worker_checkpoint_manager.recover().full_df
    # Get the workunits
    if len(full_worker_df) == 0:
        processed_reference_networks = []
        processed_target_networks = []
    else:
        processed_reference_networks  = full_worker_df[cn.FINDER_REFERENCE_NAME]
        processed_target_networks = full_worker_df[cn.FINDER_TARGET_NAME]
    workunit_manager = SubnetFinderWorkunitManager(workunit_path, num_worker=num_worker)
    workunits = workunit_manager.getWorkunits(worker_idx,
          processed_reference_networks=processed_reference_networks,
          processed_target_networks=processed_target_networks)
    # Get the reference and target models
    reference_networks = getNetworks(reference_serialization_path)
    target_networks = getNetworks(target_serialization_path)
    # Process the unprocessed reference models in batches
    msg = f"**Worker start {worker_idx} with"
    msg += " {len(reference_networks) - len(processed_reference_networks)} networks."
    if is_report_worker:
        print(msg, flush=True)
    prior_merged_checkpoint_result = None
    printPerformanceInformation(f"**Worker {worker_idx} start.")
    workunit_idx = 0
    last_checkpoint_time = time.process_time()
    for idx, workunit in enumerate(workunits):
        is_report_loop = is_report_worker and workunit_idx % REPORT_INTERVAL == 0
        finder = SubnetFinder.makeFromCombinations([reference_networks[workunit.reference_idx]],
              [target_networks[workunit.target_idx]],
              identity=identity,
              num_process=1)
        incremental_df = finder.find(is_report=is_report_loop, max_num_assignment=max_num_assignment)
        full_worker_df = pd.concat([full_worker_df, incremental_df], ignore_index=True)
        if time.process_time() - last_checkpoint_time > CHECKPOINT_TIME:
            worker_checkpoint_manager.checkpoint(full_worker_df)
            last_checkpoint_time = time.process_time()
        # Update the processed networks
        if worker_idx == 0 and workunit_idx % checkpoint_interval == 0:
            merged_checkpoint_result = WorkerCheckpointManager.merge(checkpoint_path, num_worker,
                merged_checkpoint_result=prior_merged_checkpoint_result, is_report=is_report_worker)
            prior_merged_checkpoint_result = merged_checkpoint_result
        if is_report_loop:
            print(f"**Processed {idx} work units.", flush=True)
        workunit_idx += 1
    # Do final checkpoint for worker
    worker_checkpoint_manager.checkpoint(full_worker_df)
    printPerformanceInformation(f"Completed process {worker_idx}.")
    #
    if is_report_worker:
        print(f"**Worker {worker_idx} done.", flush=True)
    if is_allow_exit:
        sys.exit(0)
    
#############################################################
class Workunit(object):
    def __init__(self, reference_idx:int=-1, target_idx:int=-1, is_done:bool=False):
        """
        Args:
            reference_idx (int, optional): index of reference network. Defaults to -1 means all.
            target_idx (int, optional): index of target network. Defaults to -1 means all.
            is_done (bool, optional): No more work.
        """
        self.reference_idx = reference_idx
        self.target_idx = target_idx
        self.is_done = is_done

    def __repr__(self):
        return f"Workunit(ref={self.reference_idx}, tgt={self.target_idx}, is_done={self.is_done})"
    
    def __eq__(self, other):
        return (self.reference_idx == other.reference_idx) and (self.target_idx == other.target_idx)
    
    @classmethod
    def addMultipleWorkunits(cls, queue, reference_idxs:Optional[List[int]]=None,
          target_idxs:Optional[List[int]]=None)->None:
        """
        Adds a list of integer workunits to the queue that is the cross product
        of the reference and target indexes.

        Args:
            queue (Queue): Queue to add the workunits
            reference_idxs (Optional[List[int]]): List of reference indexes
            target_idxs (Optional[List[int]]): List of target indexes
        """
        if reference_idxs is None:
            reference_idxs = [-1]
        if target_idxs is None:
            target_idxs = [-1]
        workunits = [cls(reference_idx=ref_idx, target_idx=tgt_idx)
              for ref_idx in reference_idxs for tgt_idx in target_idxs]
        [queue.put(w) for w in workunits]


'''Provides diagnostics for the compatibility collection result.'''


from pySubnetSB.species_constraint import SpeciesConstraint # type: ignore
from pySubnetSB.reaction_constraint import ReactionConstraint # type: ignore
from pySubnetSB.named_matrix import NamedMatrix, NULL_NMAT # type: ignore

import collections
import numpy as np # type: ignore
import pandas as pd # type: ignore

TARGET = "target"
REFERENCE = "reference"
EQUALITY = "equality"
INEQUALITY_NUMERICAL = "inequality_numerical"
INEQUALITY_BITWISE = "inequality_bitwise"
CONSTRAINT_TYPES = [EQUALITY, INEQUALITY_NUMERICAL, INEQUALITY_BITWISE]


ColumnDescriptor = collections.namedtuple("ColumnDescriptor", ["column_name", "named_matrix", "constraint_type"])


#####################################
class CompatibilityCollectionInspector(object):

    def __init__(self, reference_network, target_network, is_species:bool=True, is_subnet:bool=False):
        """
        Args:
            reference_network (Network)
            target_network (Network)
            is_species (bool, optional): _description_. Defaults to True.
            is_subnet (bool, optional): check for su
        """
        self.reference_network = reference_network
        self.target_network = target_network
        self.is_species = is_species
        self.is_subnet = is_subnet
        #
        if self.is_species:
            constraint_cls = SpeciesConstraint
        else:
            constraint_cls = ReactionConstraint
        self.reference_constraint = constraint_cls(self.reference_network.reactant_nmat,
              self.reference_network.product_nmat, is_subnet=self.is_subnet)
        self.target_constraint = constraint_cls(self.target_network.reactant_nmat,
              self.target_network.product_nmat, is_subnet=self.is_subnet)
        self.compatibility_collection_result = self.reference_constraint.makeCompatibilityCollection(
              self.target_constraint, is_diagnostic=True)
        # 
        self.column_name_dct = self._makeColumnNameDct()
        self._diagnostic_nmat = NULL_NMAT  # Deferred execution to improve testing _labelReferenceTarget
        self._diagnostic_label_nmat = NULL_NMAT  # Deferred execution to improve testing _labelReferenceTarget

    def _makeColumnNameDct(self)->dict:
        """Creates a dictionay column names and their associated NamedMatrix.

        Raises:
            ValueError: ValueError

        Returns: 
            dict:
                key: REFERENCE | TARGET, <column name>
                value: ColumnDescriptor
        """
        dct = {}
        for column_name in self.compatibility_collection_result.reference_equality_nmat.column_names:
            dct[(REFERENCE, column_name)] = ColumnDescriptor(
                  column_name=column_name,
                  named_matrix=self.compatibility_collection_result.reference_equality_nmat,
                  constraint_type=EQUALITY)
        for column_name in self.compatibility_collection_result.reference_numerical_inequality_nmat.column_names:
            dct[(REFERENCE, column_name)] = ColumnDescriptor(
                  column_name=column_name,
                  named_matrix=self.compatibility_collection_result.reference_numerical_inequality_nmat,
                  constraint_type=INEQUALITY_NUMERICAL)
        for column_name in self.compatibility_collection_result.reference_bitwise_inequality_nmat.column_names:
            dct[(REFERENCE, column_name)] = ColumnDescriptor(
                  column_name=column_name,
                  named_matrix=self.compatibility_collection_result.reference_bitwise_inequality_nmat,
                  constraint_type=INEQUALITY_BITWISE)
        #
        for column_name in self.compatibility_collection_result.target_equality_nmat.column_names:
            dct[(TARGET, column_name)] = ColumnDescriptor(
                  column_name=column_name,
                  named_matrix=self.compatibility_collection_result.target_equality_nmat,
                  constraint_type=EQUALITY)
        for column_name in self.compatibility_collection_result.target_numerical_inequality_nmat.column_names:
            dct[(TARGET, column_name)] = ColumnDescriptor(
                  column_name=column_name,
                  named_matrix=self.compatibility_collection_result.target_numerical_inequality_nmat,
                  constraint_type=INEQUALITY_NUMERICAL)
        for column_name in self.compatibility_collection_result.target_bitwise_inequality_nmat.column_names:
            dct[(TARGET, column_name)] = ColumnDescriptor(
                  column_name=column_name,
                  named_matrix=self.compatibility_collection_result.target_bitwise_inequality_nmat,
                  constraint_type=INEQUALITY_BITWISE)
        return dct
        
    @property
    def diagnostic_nmat(self)->NamedMatrix:
        if self._diagnostic_nmat is NULL_NMAT:
            self._diagnostic_nmat = self._makeDiagnosticMatrix()
        return self._diagnostic_nmat
    
    @property
    def diagnostic_label_nmat(self)->NamedMatrix:
        if self._diagnostic_label_nmat is NULL_NMAT:
            self._diagnostic_label_nmat = self._makeDiagnosticLabelMatrix()
        return self._diagnostic_label_nmat
    
    @property
    def labelled_diagnostic_nmat(self)->NamedMatrix:
        return NamedMatrix.hstack([self.diagnostic_label_nmat, self.diagnostic_nmat])

    def __bool__(self):
        # Returns True if the diagnostic matrix is not NULL_NMAT.
        return self.diagnostic_nmat is not NULL_NMAT
    
    def _makeDiagnosticLabelMatrix(self)->NamedMatrix:
        """
        Construct labels rows in the cross product of reference and target constraint matrices.
        Reference indices run slower than target, as indicated below for a comparison of species names
        with 2 species (S*) in self and 3 in other (A*).

              referencex  target
              S0           A0
              S0           A1
              S0           A2
              S1           A0
              S1           A1
              S1           A2

        Args:
            target: Constraint
            is_species:boolean (default: True) True if species, False if reactions
        """
        # Get the names of the elements being constrained
        reference_names = self.reference_constraint.row_names
        target_names = self.target_constraint.row_names
        reference_length = len(reference_names)
        target_length = len(target_names)
        # Construct the NamedMatrix
        reference_labels = np.repeat(reference_names, target_length)
        length = reference_labels.shape[0]
        reference_labels = np.reshape(reference_labels, (length, 1))
        target_labels = np.array([target_names]*reference_length).flatten()
        target_labels = np.reshape(target_labels, (length, 1))
        arr = np.concatenate([reference_labels, target_labels], axis=1)
        label_nmat = NamedMatrix(arr, column_names=["reference", "target"])
        return label_nmat
    
    def _makeDiagnosticMatrix(self)->NamedMatrix:
        # Calculates the diagnostic matrix.
        diagnostic_nmat = NULL_NMAT
        for nmat in [self.compatibility_collection_result.equality_compatibility_nmat,
                     self.compatibility_collection_result.inequality_compatibility_numerical_nmat,
              self.compatibility_collection_result.inequality_compatibility_bitwise_nmat]:
            if diagnostic_nmat == NULL_NMAT:
                diagnostic_nmat = nmat
            else:
                if nmat != NULL_NMAT:
                    diagnostic_nmat = NamedMatrix.hstack([diagnostic_nmat, nmat])
        return diagnostic_nmat
    
    def getTrueRowsInDiagnosticMatrix(self)->np.ndarray:
        """Returns the indices of the rows in the diagnostic matrix that are True.

        Returns:
            np.ndarray: Indices of the rows in the diagnostic matrix that are True.
        """
        row_idxs = [n for n in range(self.diagnostic_nmat.num_row) if np.all(self.diagnostic_nmat.values[n, :])]
        return np.array(row_idxs)

    def _getConstraintValue(self, element_name:str, column_name:str, is_reference:bool=True)->float:
        """Returns the values of the constraint

        Args:
            element_name (str): Name of the element (species or reaction).
            column_name (str): Name of the constraint.
            is_reference (bool, optional): True if the reference network. Defaults to True.

        Returns:
            np.ndarray: Column for the NamedMatrix
        """
        if is_reference:
            full_nmat = self.column_name_dct[(REFERENCE, column_name)].named_matrix
        else:
            full_nmat = self.column_name_dct[(TARGET, column_name)].named_matrix
        nmat = full_nmat.getSubNamedMatrix(row_names=[element_name], column_names=[column_name]).named_matrix
        return nmat.values.flatten()[0]
    
    def explainNotCompatible(self, reference_name:str, target_name:str)->pd.DataFrame:
        """Provides information as to why the target_name is not compatible with the reference_name.

        Args:
            reference_name (str)
            target_name (str)

        Returns:
            pd.DataFrame: Indexed by name of constraint
                reference_value: reference value
                target_value: target value
                attrbute_type: result(T/F), eq, ineq_num, ineq_bit
                comparison_result: comparison result (bool)
        """
        REFERENCE_VALUE = "reference_value"
        TARGET_VALUE = "target_value"
        CONSTRAINT_TYPE = "constraint_type"
        COMPARISON_RESULT = "comparison_result"
        # Returns a string representation of the diagnostic matrix.
        reference_target_idxs = [n for n in range(self.diagnostic_label_nmat.num_row)
              if self.diagnostic_label_nmat.values[n, 0] == reference_name and
              self.diagnostic_label_nmat.values[n, 1] == target_name]
        if len(reference_target_idxs) != 1:
            raise ValueError("No diagnostic information for the reference and target names.")
        reference_target_idx = reference_target_idxs[0]
        ser = pd.Series(self.diagnostic_nmat.values[reference_target_idx, :],
                index=self.diagnostic_nmat.column_names)
        false_constraint_column_names = ser.index[ser == False]
        # Construct the DataFrame
        dataframe_column_names = [REFERENCE_VALUE, TARGET_VALUE, CONSTRAINT_TYPE, COMPARISON_RESULT]
        dct:dict = {c: [] for c in dataframe_column_names}
        for false_constraint_column_name in false_constraint_column_names:
            false_constraint_column_idx = self.diagnostic_nmat.getColumnIdx(false_constraint_column_name)
            comparison_result = self.diagnostic_nmat.values[reference_target_idx, false_constraint_column_idx]
            reference_value = self._getConstraintValue(reference_name, false_constraint_column_name,
                  is_reference=True)
            target_value = self._getConstraintValue(target_name, false_constraint_column_name,
                  is_reference=False)
            constraint_type = self.column_name_dct[(REFERENCE, false_constraint_column_name)].constraint_type
            dct[REFERENCE_VALUE].append(reference_value)
            dct[TARGET_VALUE].append(target_value)
            dct[CONSTRAINT_TYPE].append(constraint_type)
            dct[COMPARISON_RESULT].append(comparison_result)
        df = pd.DataFrame(dct, index=false_constraint_column_names)
        return df


import pySubnetSB.util as util # type: ignore

import os
import time
import numpy as np
import unittest
from functools import cmp_to_key


IGNORE_TEST = False
IS_PLOT = False

#############################
# Tests
#############################
class TestFunctions(unittest.TestCase):

    def testTimeit(self):
        if IGNORE_TEST:
            return
        util.IS_TIMEIT = IGNORE_TEST
        @util.timeit
        def test():
            time.sleep(1)
        test()
        util.IS_TIMEIT = False

    def testRepeatRow(self):
        if IGNORE_TEST:
            return
        arr = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
        num_repeat = 2
        result = util.repeatRow(arr, num_repeat)
        for idx in range(arr.shape[0]):
            result_idx1 = num_repeat*idx
            result_idx2 = num_repeat*idx + 1
            self.assertTrue(np.all(result[result_idx1] == arr[idx]))
            self.assertTrue(np.all(result[result_idx2] == arr[idx]))

    def testRepeatArray(self):
        if IGNORE_TEST:
            return
        arr = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
        num_repeat = 2
        num_row = arr.shape[0]
        result = util.repeatArray(arr, num_repeat)
        for idx in range(num_repeat):
            rows = range(idx*num_row, (idx+1)*num_row)
            self.assertTrue(np.all(result[rows] == arr))

    def testHashArray(self):
        if IGNORE_TEST:
            return
        def test(size=20, ndim=2, num_iteration=100):
            for _ in range(num_iteration):
                row_perm = np.random.permutation(size)
                mat_perm = np.random.permutation(size)
                if ndim == 1:
                    arr1 = np.random.randint(-2, 3, size)
                    arr2 = arr1.copy()
                if ndim == 2:
                    arr1 = np.random.randint(-2, 3, (size, size))
                    arr2 = arr1.copy()
                    arr2 = arr2[row_perm, :]
                if ndim == 3:
                    arr1 = np.random.randint(-2, 3, (size, size, size))
                    arr2 = arr1.copy()
                    arr2 = arr2[mat_perm, :, :]
                    arr2 = arr2[:, row_perm, :]
                result1 = util.makeRowOrderIndependentHash(arr1)
                result2 = util.makeRowOrderIndependentHash(arr2)
                self.assertTrue(result1 == result2)
        #
        test(ndim=1)
        test(ndim=2)
        with self.assertRaises(ValueError):
            test(ndim=3)

    def testHashArrayScale(self):
        if IGNORE_TEST:
            return
        def test(size=20, num_iteration=100):
            for _ in range(num_iteration):
                row_perm = np.random.permutation(size)
                arr1 = np.random.randint(-2, 3, (size, size))
                arr2 = arr1.copy()
                arr2 = arr2[row_perm, :]
                result1 = util.makeRowOrderIndependentHash(arr1)
                result2 = util.makeRowOrderIndependentHash(arr2)
                self.assertTrue(result1 == result2)
        test(num_iteration=10000)
    
    def testHashMarixPermutations(self):
        if IGNORE_TEST:
            return
        def test(size=50, num_iteration=100):
            for _ in range(num_iteration):
                row_perm = np.random.permutation(size)
                arr1 = np.random.randint(-2, 3, (size, 2*size))
                arr2 = arr1.copy()
                arr2 = arr2[row_perm, :]
                result1 = util.hashMatrix(arr1)
                result2 = util.hashMatrix(arr2)
                self.assertTrue(result1 == result2)
        test(num_iteration=10000)

    def testHashMarixDistinct(self):
        if IGNORE_TEST:
            return
        def test(size=10, num_iteration=100):
            num_true = 0
            for _ in range(num_iteration):
                arr1 = np.random.randint(-2, 3, (size, 2*size))
                arr2 = np.random.randint(-2, 3, (size, 2*size))
                result1 = util.hashMatrix(arr1)
                result2 = util.hashMatrix(arr2)
                num_true += result1 == result2
            self.assertLess(num_true/num_iteration, 0.1)
            #print(f"{size}: fraction of true: {num_true/num_iteration}")
        test(size=3, num_iteration=1000)
        test(size=5, num_iteration=1000)
    
    def testIsLessThan(self):
        if IGNORE_TEST:
            return
        arr1 = np.array([1, 2, 3])
        arr2 = np.array([1, 2, 4])
        arr3 = np.array([1, 2, 4 , 5])
        self.assertTrue(util.isArrayLessEqual(arr1, arr2))
        self.assertFalse(util.isArrayLessEqual(arr2, arr1))
        self.assertFalse(util.isArrayLessEqual(arr1, arr3))

    def testArrayToStr(self):
        if IGNORE_TEST:
            return
        for _ in range(10):
            big_array = np.random.randint(0, 10, 100)
            big_array = np.reshape(big_array, (10, 10))
            context = util.array2Context(big_array)
            other_array = util.string2Array(context)
            self.assertTrue(np.all(big_array == other_array))
    
    def testArrayToStr2(self):
        if IGNORE_TEST:
            return
        for _ in range(10):
            big_array = np.random.randint(0, 10, 100)
            context = util.array2Context(big_array)
            other_array = util.string2Array(context)
            self.assertTrue(np.all(big_array == other_array.flatten()))

    def testPruneArray(self):
        if IGNORE_TEST:
            return
        def test(size=20, max_size=10, num_iteration=100):
            for _ in range(num_iteration):
                arr = np.random.randint(0, 3, (size, size))
                pruned_arr, is_pruned = util.pruneArray(arr, max_size)
                self.assertTrue(is_pruned)
                self.assertTrue(pruned_arr.shape[0] <= max_size)
        #
        test(size=5, max_size=2)
        test()

    def testHashMatrixTrue(self):
        if IGNORE_TEST:
            return
        def test(size=10, num_iteration=100):
            for _ in range(num_iteration):
                row_perm = np.random.permutation(size)
                arr1 = np.random.randint(0, 3, (size, size))
                arr2 = arr1.copy()
                arr2 = arr2[row_perm, :]
                result1 = util.makeRowOrderIndependentHash(arr1)
                result2 = util.makeRowOrderIndependentHash(arr2)
                self.assertTrue(result1 == result2)
        test(num_iteration=10000)

    def testHashMatrixFalse(self):
        if IGNORE_TEST:
            return
        def test(size=3, num_iteration=100):
            num_collision = 0
            for _ in range(num_iteration):
                arr1 = np.random.randint(0, 3, (size, size))
                arr2 = np.random.randint(0, 3, (size, size))
                result1 = util.makeRowOrderIndependentHash(arr1)
                result2 = util.makeRowOrderIndependentHash(arr2)
                num_collision += result1 == result2
        test(num_iteration=1000)

    def testListOfLists(self):
        if IGNORE_TEST:
            return
        for num_list in range(2, 20):
            lengths = np.random.randint(2, 10, num_list)
            list_of_lists = [np.random.randint(0, length, length) for length in lengths]
            sample_arr = util.sampleListOfLists(list_of_lists, 10)
            for sample in sample_arr:
                self.assertTrue(np.all([s in list_of_lists[i] for i, s in enumerate(sample)]))

    def testPartitionArray(self):
        if IGNORE_TEST:
            return
        arr = np.random.randint(0, 10, 45)
        arr = np.reshape(arr, (15, 3))
        num_partition = 4
        partitions = util.partitionArray(arr, num_partition)
        self.assertEqual(len(partitions), num_partition)
        lengths = [len(partition) for partition in partitions]
        self.assertTrue(np.max(lengths) - np.min(lengths) <= 1)
        for row in arr:
            is_true = False
            for partition in partitions:
                for partition_row in partition:
                    is_true = is_true or np.all(row == partition_row)
            self.assertTrue(is_true)
    
    def testPartitionArray2(self):
        if IGNORE_TEST:
            return
        arr = [ np.array([1, 0])]
        num_partition = 2
        partitions = util.partitionArray(arr, num_partition)
        self.assertEqual(len(partitions), num_partition-1)

    def testEncodeIntPairDecodeIntPair(self):
        if IGNORE_TEST:
            return
        for _ in range(100):
            int1, int2 = (np.random.randint(1000, 100000), np.random.randint(1000, 100000))
            encoded_int = util.encodeIntPair(int1, int2)
            decoded_int1, decoded_int2 = util.decodeIntPair(encoded_int)
            self.assertEqual(int1, decoded_int1)
            self.assertEqual(int2, decoded_int2)
    
    def testEncodeIntPairDecodeIntPairArray(self):
        if IGNORE_TEST:
            return
        size = 10
        for _ in range(100):
            int1 = np.random.randint(1000, 100000, size)
            int2 = np.random.randint(1000, 100000, size)
            encoded_int = util.encodeIntPair(int1, int2)
            decoded_int1, decoded_int2 = util.decodeIntPair(encoded_int)
            self.assertTrue(np.all(int1 == decoded_int1))
            self.assertTrue(np.all(int2 == decoded_int2))

    def testGetAllSubsets(self):
        if IGNORE_TEST:
            return
        arr = [1, 2, 3, 4]
        subsets = util.getAllSubsets(arr)
        self.assertEqual(len(subsets), 2**len(arr))

    def testMakeLocalFileFromURL(self):
        if IGNORE_TEST:
            return
        url = "http://raw.githubusercontent.com/ModelEngineering/pySubnetSB/main/examples/target_serialized.txt"
        local_file = util.makeLocalFileFromURL(url)
        self.assertTrue(os.path.isfile(local_file))
        with open(local_file, "r") as f:
            lines = f.readlines()
        self.assertGreater(len(lines), 1)

    def testGetStringsFromURL(self):
        if IGNORE_TEST:
            return
        url = "http://raw.githubusercontent.com/ModelEngineering/pySubnetSB/main/examples/target_serialized.txt"
        strings = util.getStringsFromURL(url)
        self.assertGreater(len(strings), 1)


if __name__ == '__main__':
    unittest.main()


'''Pairs of species and reaction assignments.'''

import pySubnetSB.constants as cn  # type: ignore
import pySubnetSB.util as util  # type: ignore

import json
import numpy as np  # type: ignore
from typing import List, Optional

class AssignmentPair(object):
    # species and row are equivalent. reaction and column are equivalent.

    def __init__(self, species_assignment=None, reaction_assignment=None, row_assignment=None, column_assignment=None,
          reference_species_names:Optional[List[str]]=None,
          target_species_names:Optional[List[str]]=None,
          reference_reaction_names:Optional[List[str]]=None,
          target_reaction_names:Optional[List[str]]=None):
        """
        Assignment of species and reactions to target to describe its permutation into the reference.

        Args:
            species_assignment (_type_, optional): Assignment of target species (values) to reference (position)
            reaction_assignment (_type_, optional): Assignment of target reactions (values) to reference (position)
            row_assignment (_type_, optional): Same as species
            column_assignment (_type_, optional): Same as reactions
            reference_species_names (Optional[List[str]], optional)
            target_species_names (Optional[List[str]], optional)
            reference_reaction_names (Optional[List[str]], optional)
            target_reaction_names (Optional[List[str]], optional)

        Raises:
            RuntimeError: _description_
            RuntimeError: _description_
        """
        if (species_assignment is None) and (row_assignment is None):
            raise RuntimeError("Must specify species assignment!")
        if (reaction_assignment is None) and (column_assignment is None):
            raise RuntimeError("Must specify reaction assignment!")
        self._species_assignment = species_assignment
        self._reaction_assignment = reaction_assignment
        self._row_assignment = row_assignment
        self._column_assignment = column_assignment
        self.reference_species_names = reference_species_names
        self.reference_reaction_names = reference_reaction_names
        self.target_species_names = target_species_names
        self.target_reaction_names = target_reaction_names

    @property
    def species_assignment(self)->np.ndarray:
        if self._species_assignment is None:
            self._species_assignment = self._row_assignment
        return self._species_assignment
    
    @property
    def row_assignment(self)->np.ndarray:
        if self._row_assignment is None:
            self._row_assignment = self._species_assignment
        return self._row_assignment
    
    @property
    def reaction_assignment(self)->np.ndarray:
        if self._reaction_assignment is None:
            self._reaction_assignment = self._column_assignment
        return self._reaction_assignment
    
    @property
    def column_assignment(self)->np.ndarray:
        if self._column_assignment is None:
            self._column_assignment = self._reaction_assignment
        return self._column_assignment

    def __repr__(self):
        return f"species: {self.species_assignment}, reaction: {self.reaction_assignment}"

    def copy(self)->'AssignmentPair':
        return AssignmentPair(species_assignment=self.species_assignment.copy(),
                              reaction_assignment=self.reaction_assignment.copy())
    
    def invert(self)->'AssignmentPair':
        """Invert the assignment pair.

        Returns:
            AssignmentPair
        """
        return AssignmentPair(species_assignment=np.argsort(self.species_assignment.copy()),
                              reaction_assignment=np.argsort(self.reaction_assignment.copy()))
    
    def __eq__(self, other)->bool:
        if not self.__class__.__name__ in str(type(other)):
            return False
        if (len(self.species_assignment) != len(other.species_assignment)):
            return False
        if (len(self.reaction_assignment) != len(other.reaction_assignment)):
            return False
        if not np.all(self.species_assignment == other.species_assignment):
            return False
        if not np.all(self.reaction_assignment == other.reaction_assignment):
            return False
        return True
    
    def serialize(self)->str:
        """Create a JSON string for the object.

        Returns:
            str: _description_
        """
        species_assignment_lst = self.species_assignment.tolist()
        reaction_assignment_lst = self.reaction_assignment.tolist()
        return json.dumps({cn.S_ID: self.__class__.__name__,
                           cn.S_SPECIES_ASSIGNMENT_LST: species_assignment_lst,
                           cn.S_REACTION_ASSIGNMENT_LST: reaction_assignment_lst})
    
    @classmethod
    def deserialize(cls, serialization_str:str)->'AssignmentPair':
        """Creates an AssignmentPair from its JSON string serialization:

        Args:
            serialization_str (str)

        Returns:
            AssignmentPair
        """
        dct = json.loads(serialization_str)
        if not cls.__name__ in dct[cn.S_ID]:
            raise ValueError(f"Expected {cls.__name__} but got {dct[cn.S_ID]}")
        species_assignment = np.array(dct[cn.S_SPECIES_ASSIGNMENT_LST])
        reaction_assignment = np.array(dct[cn.S_REACTION_ASSIGNMENT_LST])
        return AssignmentPair(species_assignment=species_assignment, reaction_assignment=reaction_assignment)
    
    def resize(self, new_size)->'AssignmentPair':
        """Truncates the assignment pair to the specify size.

        Args:
            new_size (int): New size for the assignment pair

        Returns:

        """
        return AssignmentPair(species_assignment=self.species_assignment[:new_size],
                              reaction_assignment=self.reaction_assignment[:new_size])

    def makeNameDct(self)->dict:
        """Create a dictionary of names for the assignment of target species and reactions to the reference network.

        Returns:
            dict:
              keys: cn.SPECIES_NAMES, cn.REACTION_NAMES
              values: dict
                keys: reference name
                values: target name
        """
        if (self.reference_species_names is None) or (self.target_species_names is None):
            raise ValueError("Must specify species names.")
        if (self.reference_reaction_names is None) or (self.target_reaction_names is None):
            raise ValueError("Must specify reaction names.")
        species_dct = {self.reference_species_names[i]: self.target_species_names[self.species_assignment[i]]
                       for i in range(len(self.species_assignment))}
        reaction_dct = {self.reference_reaction_names[i]: self.target_reaction_names[self.reaction_assignment[i]]
                       for i in range(len(self.reaction_assignment))}
        return {cn.SPECIES_NAMES: species_dct, cn.REACTION_NAMES: reaction_dct}


import pySubnetSB.constants as cn  # type: ignore
from pySubnetSB.subnet_finder_workunit_manager import SubnetFinderWorkunitManager # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.model_serializer import ModelSerializer  # type: ignore

import os
import numpy as np
import tellurium as te # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
WORKUNIT_PATH = os.path.join(cn.TEST_DIR, "test_subnet_finder_workunit_manager.csv")
TARGET_SERIALIZATION_PATH = os.path.join(cn.TEST_DIR, "test_subnet_finder_workunit_manager_targets.txt")
REFERENCE_SERIALIZATION_PATH = os.path.join(cn.TEST_DIR,
      "test_subnet_finder_workunit_manager_references.txt")
NUM_REFERENCE_NETWORK = 5
NUM_TARGET_NETWORK = 6
NUM_WORKER = 10
REFERENCE_NETWORKS = [Network.makeRandomNetworkByReactionType(10, is_prune_species=True)
      for _ in range(NUM_REFERENCE_NETWORK)]
TARGET_NETWORKS = [Network.makeRandomNetworkByReactionType(10, is_prune_species=True)
      for _ in range(NUM_TARGET_NETWORK)]


#############################
# Tests
#############################
class TestSubnetFinderWorkunitManager(unittest.TestCase):

    def setUp(self):
        self.manager = SubnetFinderWorkunitManager(WORKUNIT_PATH,
              num_worker=NUM_WORKER, reference_networks=REFERENCE_NETWORKS,
                target_networks=TARGET_NETWORKS)
        self.num_workunit = NUM_REFERENCE_NETWORK * NUM_TARGET_NETWORK
        self.workunit_per_worker = self.num_workunit / NUM_WORKER
        self.manager.remove()

    def tearDown(self):
        self.manager.remove()
        for serialization_path in [TARGET_SERIALIZATION_PATH, REFERENCE_SERIALIZATION_PATH]:
            serializer = ModelSerializer(None, serialization_path)
            serializer.remove()

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertEqual(NUM_WORKER, self.manager.num_worker)
        self.assertEqual(NUM_REFERENCE_NETWORK, self.manager.num_reference_network)
        self.assertEqual(NUM_TARGET_NETWORK, self.manager.num_target_network)
        self.assertEqual(WORKUNIT_PATH, self.manager.csv_path)

    def testMakeWorkunitFile(self):
        if IGNORE_TEST:
            return
        serializer = ModelSerializer(None, serialization_path=REFERENCE_SERIALIZATION_PATH)
        serializer.serializeNetworks(REFERENCE_NETWORKS)
        serializer = ModelSerializer(None, serialization_path=TARGET_SERIALIZATION_PATH)
        serializer.serializeNetworks(TARGET_NETWORKS)
        self.manager.makeWorkunitFile()
        df = self.manager.getWorkunitDataframe()
        self.assertEqual(NUM_REFERENCE_NETWORK*NUM_TARGET_NETWORK, len(df))
        for worker_idx in range(NUM_WORKER):
            count = len(df[df[cn.FINDER_WORKER_IDX] == worker_idx])
            self.assertTrue(count >= self.workunit_per_worker-1)
    
    def testMakeWorkunitFileFromPath(self):
        if IGNORE_TEST:
            return
        self.manager.makeWorkunitFile()
        df = self.manager.getWorkunitDataframe()
        self.assertEqual(NUM_REFERENCE_NETWORK*NUM_TARGET_NETWORK, len(df))
        for worker_idx in range(NUM_WORKER):
            count = len(df[df[cn.FINDER_WORKER_IDX] == worker_idx])
            self.assertTrue(count >= self.workunit_per_worker-1)

    def testMakeWorkunitFile2(self):
        if IGNORE_TEST:
            return
        self.manager.makeWorkunitFile()
        worker_df = self.manager.getWorkunitDataframe()
        # Ensure that all workunits are present and unique
        for reference_idx in range(NUM_REFERENCE_NETWORK):
            for target_idx in range(NUM_TARGET_NETWORK):
                workunit = worker_df[(worker_df[cn.FINDER_REFERENCE_IDX] == reference_idx) &
                      (worker_df[cn.FINDER_TARGET_IDX] == target_idx)]
                self.assertTrue(len(workunit) == 1)

    def testGetWorkunits(self):
        if IGNORE_TEST:
            return
        self.manager.makeWorkunitFile()
        for worker_idx in range(NUM_WORKER):
            workunits = self.manager.getWorkunits(worker_idx)
            self.assertTrue(np.abs(self.workunit_per_worker - len(workunits)) <= 1)

    def testGetWorkunitsPriorWork(self):
        if IGNORE_TEST:
            return
        num_processed = 2   # Delete some workunits. They should be recovered.
        worker_idx = 0
        self.manager.makeWorkunitFile()
        worker_df = self.manager.getWorkunitDataframe(worker_idx=worker_idx)
        processed_reference_networks = [worker_df.iloc[i][cn.FINDER_REFERENCE_NAME]
              for i in range(num_processed)]   
        processed_target_networks = [worker_df.iloc[i][cn.FINDER_TARGET_NAME]
              for i in range(num_processed)]   
        # Delete some workunits
        # Should just have the deleted workunits
        workunits = self.manager.getWorkunits(worker_idx,
              processed_target_networks=processed_target_networks,
              processed_reference_networks=processed_reference_networks)
        self.assertEqual(len(workunits), len(worker_df) - num_processed)

    def testRemove(self):
        if IGNORE_TEST:
            return
        # Smoke test
        self.manager.remove()
        # Create file
        self.manager.makeWorkunitFile()
        self.manager.remove()
        self.assertFalse(os.path.exists(WORKUNIT_PATH))


if __name__ == '__main__':
    unittest.main()


'''Builds ProcessedNetworks from a NetworkCollection based on their structural identity.'''

from pySubnetSB import constants as cn  # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.network_collection import NetworkCollection  # type: ignore
from pySubnetSB.processed_network import ProcessedNetwork # type: ignore
from pySubnetSB.processed_network_collection import ProcessedNetworkCollection # type: ignore

import numpy as np
import pandas as pd  # type: ignore
import time
from typing import List, Dict


###############################################
class ClusterBuilder(object):
    # Builds ClusterNetworks from a NetworkCollection based on their structural identity

    def __init__(self, network_collection:NetworkCollection, is_report=True,
                 max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT, is_sirn:bool=True,
                 identity:str=cn.ID_WEAK):
        """
        Args:
            network_collection (NetworkCollection): Collection of networks to cluster
            is_report (bool, optional): Progress reporting
            max_num_perm (float, optional): Maximum log10 of the number of permutations that
                are examined
            is_sirn (bool, optional): Whether the SIRN algorithm is used
            is_structural_identity_strong (bool, optional): Criteria for structurally identical
        """
        self.network_collection = network_collection
        self.is_sirn = is_sirn
        self.is_report = is_report # Progress reporting
        self.max_num_assignment = max_num_assignment  # Maximum number of assignments permitted before indeterminate
        self.identity = identity
        # Statistics
        self.hash_dct = self._makeHashDct()
        self.num_hash = len(self.hash_dct)
        self.max_hash = max([len(l) for l in self.hash_dct.values()])  # Most NetworkCollection
        # Results
        self.processed_network_collections:List[ProcessedNetworkCollection] = []

    @property
    def num_indeterminant(self)->int:
        count = 0
        for processed_network_collection in self.processed_network_collections:
            count += sum([cn.is_indeterminate for cn in processed_network_collection.processed_networks])
        return count
    
    @property
    def collection_sizes(self)->List[int]:
        return [len(cnc) for cnc in self.processed_network_collections]

    @staticmethod
    def sequenceMax(sequence:List[int])->int:
        if len(sequence) == 0:
            return 0
        return max(sequence)
    
    def _makeHashDct(self)->Dict[int, List[Network]]:
        """
        Makes the hash dictionary for the network collection

        Returns:
            Dict[int, List[Network]]: _description_
        """
        def makeDct(attr:str)->Dict[int, List[Network]]:
            # Build the hash dictionary based on the attribute
            hash_dct: Dict[int, List[Network]] = {}
            # Build the hash dictionary
            for network in self.network_collection.networks:
                hash_val = getattr(network, attr)
                if hash_val in hash_dct:
                    hash_dct[hash_val].append(network)
                else:
                    hash_dct[hash_val] = [network]
            return hash_dct
        #
        if self.is_sirn:
            hash_dct = makeDct('network_hash')
        else:
            hash_dct = {cn.OTHER_HASH: self.network_collection.networks}
        return hash_dct

    def processed2Network(self, processed_network:ProcessedNetwork)->Network:
        result = self.network_collection.network_dct[processed_network.network_name]
        return result

    def cluster(self)->None:
        """
        Clusters the network in the collection by finding those that have structural identity. ProcessedNetwork
        is a wrapper for a network to provide context for clustering.

        Pseudo code:
        For all hash values
            For network with the hash value
                processed_network_collections = []
                processed_network = ProcessedNetwork(network)
                For other_processed_network in processed_network_collections
                    If the network is structurally identical to any network with this hash value
                        Add the network to a ProcessedNetworkCollection for that network

        Returns:
            Updates sef.processed_network_collections
        """
        # Initialize result
        self.processed_network_collections = []
        #print(f"\n**Number of hash values: {self.num_hash}", end="")
        # Construct collections of structurally identical Networks
        for idx, (hash_val, hash_networks) in enumerate(self.hash_dct.items()):
            if self.is_report:
                print(f" {np.round((idx+1)/self.num_hash, 2)}.", end="")
            # No processing time for the first network in a hash
            first_processed_network = ProcessedNetwork(hash_networks[0].network_name)
            # Create list of new collections for this key of hash_dct
            new_processed_network_collections =  \
                [ProcessedNetworkCollection([first_processed_network],
                     identity=self.identity,
                     hash_val=hash_val)]
            # Find structurally identical networks and add to the appropriate ProcessedNetworkCollection,
            # creating new ProcessedNetworkCollections as needed.
            for network in hash_networks[1:]:
                processed_network = ProcessedNetwork(network)  # Wrapper for clustering networks
                is_any_indeterminate = False
                is_selected = False
                start_time = time.process_time()
                for processed_network_collection in new_processed_network_collections:
                    first_processed_network = processed_network_collection.processed_networks[0]
                    first_network = self.processed2Network(first_processed_network)
                    result = first_network.isStructurallyIdentical(network,
                            max_num_assignment=self.max_num_assignment,
                            identity=self.identity, is_subnet=False, is_report=self.is_report)
                    if result:
                        processed_network_collection.add(processed_network)
                        is_selected = True
                        break
                    if (not result) and result.is_truncated:
                        is_any_indeterminate = True
                # Add statistics to the ProcessedNetwork
                processed_network.addProcessingTime(time.process_time() - start_time)
                if is_selected:
                    processed_network.setIndeterminate(False)
                    processed_network.setAssignmentCollection(result.assignment_pairs)
                else:
                    # Not structurally identical to any ProcessedNetworkCollection with this hash.
                    # Create a new ProcessedNetworkCollection for this hash.
                    processed_network.setIndeterminate(is_any_indeterminate)
                    processed_network_collection = ProcessedNetworkCollection([processed_network],
                        identity=self.identity,
                        hash_val=hash_val)
                    new_processed_network_collections.append(processed_network_collection)
            self.processed_network_collections.extend(new_processed_network_collections)
            if self.is_report:
                print(".", end='')
        if self.is_report:
            print(f"\n . Number of network collections: {len(self.processed_network_collections)}")

    def serializeProcessedNetworkCollections(self)->pd.DataFrame:
        """
        Serializes the clustering result. Information about the clustering is contained in df.attrs,
        a dict of the form {property_name: property_value}.

        Returns:
            pd.DataFrame: Serialized data
        """
        # Augment with the ProcessedNetwork information
        values = [str(v) for v in self.processed_network_collections]
        dct = {"processed_network_repr": values}
        df = pd.DataFrame(dct)
        # Augment the dataframe
        df.attrs = {cn.STRUCTURAL_IDENTITY: self.identity,
                       cn.NUM_HASH: self.num_hash,
                       cn.MAX_HASH: self.max_hash}
        return df
    
    @classmethod
    def deserializeProcessedNetworkCollections(cls, df:pd.DataFrame)->List[ProcessedNetworkCollection]:
        """
        Deserializes the clustering result.

        Args:
            df (pd.DataFrame): Serialized data

        Returns:
            ClusterBuilder: Deserialized data
        """
        return [ProcessedNetworkCollection.makeFromRepr(repr_str)
                                         for repr_str in df.processed_network_repr.values]
    
    def makeNetworkFromProcessedNetwork(self, processed_network:ProcessedNetwork)->Network:
        """
        Makes a Network from a ProcessedNetwork

        Args:
            processed_network (ProcessedNetwork): _description_

        Returns:
            Network: _description_
        """
        network_name = ProcessedNetwork.convertToNetworkName(processed_network)
        return self.network_collection.network_dct[network_name]  


from pySubnetSB.constraint_option_collection import ReactionConstraintOptionCollection # type: ignore
from pySubnetSB.constraint_option_collection import SpeciesConstraintOptionCollection  # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.species_constraint import SpeciesConstraint  # type: ignore
import pySubnetSB.constants as cn  # type: ignore

import numpy as np
from scipy.special import factorial  # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
MODEL1 = """
J0: S0 -> S1; k0*S0
J1: S1 -> S0; k1*S1
k0 = 0.1; k1 = 0.2;
S0 = 10; S1 = 20
"""
NETWORK1 = Network.makeFromAntimonyStr(MODEL1)
MODEL2 = """
J0: S0 -> S1; k0*S0
J1: S1 -> S1; k1*S1
J2: S0 + S1 -> S1; k2*S0*S1
k0 = 0.1; k1 = 0.2; k2 = 0.3
S0 = 10; S1 = 20; S2 = 30
"""
NETWORK2 = Network.makeFromAntimonyStr(MODEL2)


#############################
# Tests
#############################
class TestConstraintOptionCollection(unittest.TestCase):

    def setUp(self):
        if np.random.rand() > 0.5:
            self.option_collection = ReactionConstraintOptionCollection()
        else:
            self.option_collection = SpeciesConstraintOptionCollection()

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertTrue(all(self.option_collection.__dict__.values()))

    def testMakeFromShortName(self):
        if IGNORE_TEST:
            return
        collection_short_name = self.option_collection.collection_short_name
        new_option = self.option_collection.makeFromCollectionShortName(collection_short_name)
        self.assertEqual(self.option_collection, new_option)

    def test_sort_names(self):
        #if IGNORE_TEST:
        #    return
        short_name = self.option_collection.collection_short_name
        self.assertTrue("+" in short_name)
        #
        self.option_collection.setAllFalse()
        short_name = self.option_collection.collection_short_name
        self.assertEqual(short_name, cn.NONE)

    def testGetTrueNames(self):
        if IGNORE_TEST:
            return
        names = self.option_collection.getTrueNames()
        num_true = np.sum([self.option_collection.__dict__[k]
              for k, v in self.option_collection.__dict__.items() if isinstance(v, bool)])
        self.assertEqual(len(names), num_true)
        #
        self.option_collection.setAllFalse()
        names = self.option_collection.getTrueNames()
        self.assertEqual(len(names), 0)

    def testIterator(self):
        if IGNORE_TEST:
            return
        options = list(self.option_collection.iterator())
        self.assertEqual(len(options), 2**len(self.option_collection.option_names))
        trues = [isinstance(opt, self.option_collection.__class__) for opt in options]
        self.assertTrue(all(trues))

    def testConstraintOptionEffect(self):
        # Test if add constraints reduces number of assignments
        if IGNORE_TEST:
            return
        all_reference_constraint = SpeciesConstraint(NETWORK1.reactant_nmat, NETWORK1.product_nmat)
        all_target_constraint = SpeciesConstraint(NETWORK2.reactant_nmat, NETWORK2.product_nmat)
        all_compatibility_collection = all_reference_constraint.makeCompatibilityCollection(
            all_target_constraint).compatibility_collection
        constraint_options = SpeciesConstraintOptionCollection()
        constraint_options.setAllFalse()
        no_reference_constraint = SpeciesConstraint(NETWORK1.reactant_nmat, NETWORK1.product_nmat,
              species_constraint_option_collection=constraint_options)
        no_constraint = SpeciesConstraint(NETWORK2.reactant_nmat, NETWORK2.product_nmat,
              species_constraint_option_collection=constraint_options)
        no_compatibility_collection = no_reference_constraint.makeCompatibilityCollection(
            no_constraint).compatibility_collection
        # The compatibility set should be larger with no constraints
        trues = [len(x) > len(y) for x, y in zip(no_compatibility_collection.compatibilities,
                                                 all_compatibility_collection.compatibilities)]
        self.assertTrue(all(trues))


if __name__ == '__main__':
    unittest.main(failfast=True)


from pySubnetSB.network_collection import NetworkCollection # type: ignore
from pySubnetSB.processed_network import ProcessedNetwork # type: ignore
from pySubnetSB.processed_network_collection import ProcessedNetworkCollection # type: ignore
import pySubnetSB.constants as cn  # type: ignore

import copy
import pandas as pd # type: ignore
import numpy as np # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
HASH_VAL = 1111
COLLECTION_SIZE = 50
NETWORK_COLLECTION = NetworkCollection.makeRandomCollection(num_network=COLLECTION_SIZE)


#############################
# Tests
#############################
class TestProcessedNetworkCollection(unittest.TestCase):

    def setUp(self):
        network_collection = copy.deepcopy(NETWORK_COLLECTION.networks)
        self.processed_networks = [ProcessedNetwork(n.network_name) for n in network_collection]
        self.processed_network_collection = ProcessedNetworkCollection(self.processed_networks,
                                                                       hash_val=HASH_VAL)
        collection = NetworkCollection.makeRandomCollection(num_network=1)
        self.other_processed_network = ProcessedNetwork(collection.networks[0])

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertTrue(isinstance(self.processed_network_collection, ProcessedNetworkCollection))
        self.assertEqual(len(self.processed_network_collection), COLLECTION_SIZE)

    def testCopyEqual(self):
        if IGNORE_TEST:
            return
        copy_collection = self.processed_network_collection.copy()
        self.assertEqual(self.processed_network_collection, copy_collection)

    def testAdd(self):
        if IGNORE_TEST:
            return
        current_len = len(self.processed_network_collection)
        self.processed_network_collection.add(self.other_processed_network)
        self.assertEqual(current_len+1, len(self.processed_network_collection))
    
    def testIsSubset(self):
        if IGNORE_TEST:
            return
        processed_network_collection = self.processed_network_collection.copy()
        self.assertTrue(self.processed_network_collection.isSubset(processed_network_collection))
        #
        processed_network_collection.processed_networks =  \
              processed_network_collection.processed_networks[1:]
        self.assertFalse(self.processed_network_collection.isSubset(processed_network_collection))
    
    def testSerializeDeserialize(self):
        if IGNORE_TEST:
            return
        for identity in [cn.ID_WEAK, cn.ID_STRONG]:
            self.processed_network_collection.identity = identity
            serialization_str = self.processed_network_collection.serialize()
            processed_network_collection = self.processed_network_collection.deserialize(serialization_str)
            self.assertEqual(self.processed_network_collection, processed_network_collection)


if __name__ == '__main__':
    unittest.main()


from src.pySubnetSB.benchmark import Benchmark, C_LOG10_NUM_PERMUTATION, C_TIME  # type: ignore
from pySubnetSB.species_constraint import SpeciesConstraint  # type: ignore
from pySubnetSB.reaction_constraint import ReactionConstraint  # type: ignore
import pySubnetSB.constants as cn # type: ignore

import os
import pandas as pd # type: ignore
import itertools
import unittest


IGNORE_TEST = False
IS_PLOT = False
NUM_REACTION = 5
NUM_SPECIES = 5
FILL_SIZE = 5
NUM_ITERATION = 10


#############################
# Tests
#############################
class TestBenchmark(unittest.TestCase):

    def setUp(self):
        self.benchmark = Benchmark(NUM_REACTION, fill_size=FILL_SIZE,
              num_iteration=NUM_ITERATION)

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertEqual(self.benchmark.num_reaction, NUM_REACTION)
        self.assertEqual(len(self.benchmark.reference_networks), NUM_ITERATION)
        self.assertEqual(len(self.benchmark.target_networks), NUM_ITERATION)

    def testGetConstraintClass(self):
        if IGNORE_TEST:
            return
        constraint_class = self.benchmark._getConstraintClass(is_species=True)
        self.assertEqual(constraint_class, SpeciesConstraint)
        constraint_class = self.benchmark._getConstraintClass(is_species=False)
        self.assertEqual(constraint_class, ReactionConstraint)

    def validateBenchmarkDataframe(self, benchmark, df):
        self.assertTrue(C_TIME in df.columns)
        self.assertTrue(C_LOG10_NUM_PERMUTATION in df.columns)
        self.assertEqual(len(df), benchmark.num_iteration)

    def testRun(self):
        if IGNORE_TEST:
            return
        for is_species in [True, False]:
            for is_subnet in [True, False]:
                df = self.benchmark.run(is_species=is_species, is_subnet=is_subnet)
                self.validateBenchmarkDataframe(self.benchmark, df)

    def testRunIsContainsReferenceFalse(self):
        if IGNORE_TEST:
            return
        benchmark = Benchmark(NUM_REACTION, NUM_SPECIES, NUM_ITERATION,
              is_contains_reference=False)
        for is_species in [True, False]:
            for is_subnet in [True, False]:
                df = benchmark.run(is_species=is_species, is_subnet=is_subnet)
                self.validateBenchmarkDataframe(benchmark, df)

    def testPlotConstraintStudy(self):
        if IGNORE_TEST:
            return
        for size in range(9, 10):
            self.benchmark.plotConstraintStudy(size, size, 10, is_plot=IS_PLOT)

    def testPlotHeatmap(self):
        if IGNORE_TEST:
            return
        ax = self.benchmark.plotHeatmap(range(5, 10, 2), range(10, 30, 3), percentile=50, is_plot=IS_PLOT,
                                        num_iteration=10)
        self.assertTrue("Axes" in str(type(ax)))

    def testPlotHeatmapIscontainsFalse(self):
        if IGNORE_TEST:
            return
        ax = self.benchmark.plotHeatmap(range(5, 10, 2), range(10, 30, 3), percentile=50, is_plot=IS_PLOT,
              num_iteration=10, is_contains_reference=False)
        self.assertTrue("Axes" in str(type(ax)))

    def testPlotHeatmapNoConstraint(self):
        if IGNORE_TEST:
            return
        ax = self.benchmark.plotHeatmap(range(5, 10, 2), range(10, 30, 3), percentile=50, is_plot=IS_PLOT,
              is_no_constraint=True, num_iteration=10, title="No Constraint", num_digit=0,
              font_size=14)
        self.assertTrue("Axes" in str(type(ax)))

    def testCompareConstraints(self):
        if IGNORE_TEST:
            return
        reference_size = 3
        target_size = 8
        fill_size = target_size - reference_size
        benchmark = Benchmark(reference_size, fill_size=fill_size,
                num_iteration=NUM_ITERATION)
        for is_subnet in [True, False]:
            result = benchmark.compareConstraints(is_subnet=is_subnet)
            for dimension_result in [result.species_dimension_result, result.reaction_dimension_result]:
                self.assertTrue(isinstance(dimension_result.dataframe, pd.DataFrame))
                self.assertGreater(len(dimension_result.dataframe), 0)
            self.assertEqual(result.reference_size, reference_size)
            self.assertEqual(result.target_size, target_size)

    def testPlotCompareConstraints(self):
        if IGNORE_TEST:
            return
        reference_size = 20
        target_size = 100
        fill_size = target_size - reference_size
        benchmark = Benchmark(reference_size, fill_size=fill_size,
                num_iteration=100)
        benchmark.plotCompareConstraints(is_plot=IS_PLOT, is_subnet=True)

    def testCalculateOccurrence(self):
        if IGNORE_TEST:
            return
        pairs = [(3, 3), (3, 10), (10, 10)]
        df = Benchmark.calculateOccurrence(pairs,
              num_iteration=1000, num_replication=5, is_report=IS_PLOT)
        self.assertTrue(isinstance(df, pd.DataFrame))
        self.assertEqual(len(df), len(pairs))
        self.assertGreaterEqual(df.loc[0, cn.D_MEAN_PROBABILITY], df.loc[1, cn.D_MEAN_PROBABILITY])

    def testCalculateOccurrenceEdgeCases(self):
        if IGNORE_TEST:
            return
        pairs = [(6, 2), (2, 6), (2, 10), (10, 2)]
        df = Benchmark.calculateOccurrence(pairs,
              num_iteration=10, num_replication=5, is_report=IS_PLOT)
        self.assertTrue(isinstance(df, pd.DataFrame))
        self.assertEqual(len(df), len(pairs))
    
    def testPlotSpeciesReactionHeatmap(self):
        if IGNORE_TEST:
            return
        sizes = list(itertools.product(range(3, 10), range(3, 10)))
        done = False
        if os.path.isfile("benchmark.csv"):
            df = pd.read_csv("benchmark.csv")
            if len(df) == len(sizes):
                done = True
        if not done:
            df = Benchmark.calculateOccurrence(sizes,
               num_iteration=1000, num_replication=5)
            df.to_csv("benchmark.csv", index=False)
        Benchmark.plotSpeciesReactionHeatmap(df, cn.D_MEAN_PROBABILITY, is_plot=IS_PLOT,
              font_size=14)


if __name__ == '__main__':
    unittest.main(failfast=True)


from pySubnetSB.constraint import NULL_NMAT  # type: ignore
from pySubnetSB.reaction_constraint import ReactionConstraint # type: ignore
from pySubnetSB.named_matrix import NamedMatrix # type: ignore
from pySubnetSB.network import Network # type: ignore

import numpy as np
import re
import unittest


IGNORE_TEST = False
IS_PLOT = False
REACTANT_MATRIX = NamedMatrix(np.array([[1, 0], [0, 1], [0, 0]]))
PRODUCT_MATRIX = NamedMatrix(np.array([[1, 1], [1, 0], [0, 0]]))
#PRODUCT_MATRIX = NamedMatrix(np.array([[0, 1], [1, 0], [0, 0]]))


#############################
# Tests
#############################
class TestReactionConstraint(unittest.TestCase):

    def setUp(self):
        self.constraint = ReactionConstraint(REACTANT_MATRIX.copy(), PRODUCT_MATRIX.copy()) 

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertEqual(self.constraint.reactant_nmat, REACTANT_MATRIX)
        self.assertEqual(self.constraint.product_nmat, PRODUCT_MATRIX)
        self.assertEqual(self.constraint._numerical_categorical_nmat, NULL_NMAT)
        self.assertEqual(self.constraint._numerical_enumerated_nmat, NULL_NMAT)

    def testMakeReactionClassificationConstraintMatrix(self):
        if IGNORE_TEST:
            return
        size = 20
        network = Network.makeRandomNetworkByReactionType(size, size)
        constraint = ReactionConstraint(network.reactant_nmat, network.product_nmat)
        named_matrix = constraint._makeClassificationConstraintMatrix()
        self.assertTrue(isinstance(named_matrix, NamedMatrix))
        df = named_matrix.dataframe
        self.assertEqual(len(df), size)
    
    def testMakeSuccessorPredecessorConstraintMatrix(self):
        if IGNORE_TEST:
            return
        frac_duplicates = []
        for _ in range(100):
            size = 5
            network = Network.makeRandomNetworkByReactionType(size, size)
            constraint = ReactionConstraint(network.reactant_nmat, network.product_nmat)
            named_matrix = constraint.makeSuccessorPredecessorConstraintMatrix()
            num_unique = len(np.unique(named_matrix.values, axis=0))
            frac_duplicate = 1 - num_unique / len(named_matrix.values)
            frac_duplicates.append(frac_duplicate)
            sum_arr = named_matrix.values.sum(axis=1)
            self.assertTrue(isinstance(named_matrix, NamedMatrix))
            self.assertTrue(np.all(sum_arr >= 0))
        #print(np.mean(frac_duplicates))

    def testMakeAutocatalysisConstraintMatrix(self):
        if IGNORE_TEST:
            return
        for _ in range(100):
            size = 10  # Doesn't work for two digit species
            network = Network.makeRandomNetworkByReactionType(size, size)
            constraint = ReactionConstraint(network.reactant_nmat, network.product_nmat)
            named_matrix = constraint._makeAutocatalysisConstraintMatrix()
            self.assertTrue(isinstance(named_matrix, NamedMatrix))
            df = named_matrix.dataframe
            self.assertGreater(len(df), 0)
            # Verify the count of autocatalysis reactions
            num_autocatalysis = np.sum([len(re.findall("S" + str(n) + " .*->.*S" + str(n), str(network)))
                  for n in range(size)])
            # Regular expression counts occurrences, not reactions and so may exceed named_matrix sum.
            self.assertLessEqual(df.values.sum(), num_autocatalysis)

    def testcategoricalAndEnumeratedConstraints(self):
        if IGNORE_TEST:
            return
        for _ in range(4):
            self.constraint.setSubset(True)
            self.assertTrue(self.constraint.equality_nmat is not NULL_NMAT)
            self.assertTrue(self.constraint.numerical_inequality_nmat is not NULL_NMAT)
            #
            self.constraint.setSubset(False)
            self.assertTrue(self.constraint.equality_nmat is not NULL_NMAT)
            self.assertTrue(self.constraint.numerical_inequality_nmat is NULL_NMAT)

    def testmakeCompatibilityCollection(self):
        if IGNORE_TEST:
            return
        num_permutations = []
        for _ in range(100):
            reference_size = 18
            filler_size = 3*reference_size
            network = Network.makeRandomNetworkByReactionType(reference_size, reference_size)
            big_network = network.fill(num_fill_reaction=filler_size, num_fill_species=filler_size)
            # Not doing initialization
            reaction_constraint = ReactionConstraint(network.reactant_nmat, network.product_nmat,
                                                   is_subnet=True)
            big_reaction_constraint = ReactionConstraint(big_network.reactant_nmat, big_network.product_nmat,
                                                       is_subnet=True)
            compatibility_collection = reaction_constraint.makeCompatibilityCollection(
                  big_reaction_constraint).compatibility_collection
            name_arr = np.array(big_reaction_constraint.reactant_nmat.column_names)
            for i, arr in enumerate(compatibility_collection.compatibilities):
                reference_name = "J" + str(i)
                target_names = [name_arr[i] for i in arr]
                self.assertTrue(reference_name in target_names)
            num_permutations.append(compatibility_collection.log10_num_assignment)
        #print(np.mean(num_permutations))

    def testBug(self):
        if IGNORE_TEST:
            return
        small_model = """
        J0: S1 -> S1
        J1: S0 + S1 -> S0 + S1
        J2: S0 -> S0
        """
        network = Network.makeFromAntimonyStr(small_model)
        big_model = """
        J5: S2 -> S2 + S4
  J0: S1 -> S1
  J4: S3 -> S3
  J1: S1 + S0 -> S1 + S0
  J3:  -> S4
  J2: S0 -> S0
        """
        big_network = Network.makeFromAntimonyStr(big_model)
        reaction_constraint = ReactionConstraint(network.reactant_nmat, network.product_nmat,
                                                   is_subnet=True)
        big_reaction_constraint = ReactionConstraint(big_network.reactant_nmat, big_network.product_nmat,
                                                       is_subnet=True)
        compatibility_collection = reaction_constraint.makeCompatibilityCollection(
                  big_reaction_constraint).compatibility_collection
        name_arr = np.array(big_reaction_constraint.reactant_nmat.column_names)
        for i, arr in enumerate(compatibility_collection.compatibilities):
                reference_name = "J" + str(i)
                target_names = [name_arr[i] for i in arr]
                self.assertTrue(reference_name in target_names)

    def testBugReactionConstraint(self):
        #if IGNORE_TEST:
        #    return
        reference_model = """
            A -> B; 1
            B -> C; 1
            A = 0
            B = 0
            """
        reference_network = Network.makeFromAntimonyStr(reference_model)
        reference_constraint = ReactionConstraint(reference_network.reactant_nmat, reference_network.product_nmat,
                is_subnet=True)
        target_model = """
            S1 -> S2; 2
            S2 -> S3; 2
            S3 -> S4; 2
            A = 0
            B = 0
            C = 0
            """
        target_network = Network.makeFromAntimonyStr(target_model)
        target_constraint = ReactionConstraint(target_network.reactant_nmat, target_network.product_nmat,
                is_subnet=True)
        compatibility_collection = reference_constraint.makeCompatibilityCollection(
              target_constraint).compatibility_collection
        trues = [len(v) > 0 for v in compatibility_collection.compatibilities]
        self.assertTrue(all(trues))


if __name__ == '__main__':
    unittest.main()


'''
1. Not selecting target correctly
2. Need to clean up created files
3. Is dependency injection clean
'''

import pySubnetSB.constants as cn  # type: ignore
from pySubnetSB.parallel_subnet_finder import ParallelSubnetFinder # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.model_serializer import ModelSerializer  # type: ignore
from pySubnetSB.parallel_subnet_finder_worker import WorkerCheckpointManager  # type: ignore
import pySubnetSB.util as util # type: ignore

#util.IS_TIMEIT = True

import os
import unittest

IGNORE_TEST = False
IS_PLOT =  False
SIZE = 10
MODEL_DIR = os.path.join(cn.TEST_DIR, "oscillators")
WORKUNIT_PATH = os.path.join(cn.TEST_DIR, "test_parallel_subset_finder_workunit.csv")
BIOMODELS_DIR = os.path.join(cn.TEST_DIR, "xml_files")
BIOMODELS_SERIALIZATION_PATH = os.path.join(BIOMODELS_DIR, "biomodels_serialized.txt")
CHECKPOINT_PATH = os.path.join(cn.TEST_DIR, "test_parallel_subnet_finder_checkpoint.csv")
REFERENCE_SERIALIZATION_FILENAME = "test_parallel_subnet_finder_reference.txt"
TARGET_SERIALIZATION_FILENAME = "test_parallel_subnet_finder_target.txt"
REFERENCE_SERIALIZATION_PATH = os.path.join(cn.TEST_DIR, REFERENCE_SERIALIZATION_FILENAME)
TARGET_SERIALIZATION_PATH = os.path.join(cn.TEST_DIR, TARGET_SERIALIZATION_FILENAME)
REMOVE_FILES = [CHECKPOINT_PATH, REFERENCE_SERIALIZATION_PATH, TARGET_SERIALIZATION_PATH, WORKUNIT_PATH]
REMOVE_FILES.extend([os.path.join(cn.TEST_DIR, "test_parallel_subnet_finder_checkpoint_%d.csv" % n)
      for n in range(10)])
NUM_NETWORK = 10


#############################
class TestParallelSubnetFinder(unittest.TestCase):

    def setUp(self):
        self.remove()
        reference_networks = [Network.makeRandomNetworkByReactionType(SIZE, is_prune_species=True)
                for _ in range(NUM_NETWORK)]
        target_networks = [n.fill(num_fill_reaction=SIZE, num_fill_species=SIZE) for n in reference_networks]
        self.makeSerialization(REFERENCE_SERIALIZATION_PATH, reference_networks)
        self.makeSerialization(TARGET_SERIALIZATION_PATH, target_networks)
        self.finder = ParallelSubnetFinder(REFERENCE_SERIALIZATION_PATH,
              TARGET_SERIALIZATION_PATH, workunit_csv_path=WORKUNIT_PATH, identity = cn.ID_STRONG,
              checkpoint_path=CHECKPOINT_PATH)

    def makeSerialization(sef, path, networks):
        serializer = ModelSerializer(None, path)
        serializer.serializeNetworks(networks)

    def tearDown(self):
        self.remove()

    def remove(self):
        WorkerCheckpointManager.deleteWorkerCheckpoints(CHECKPOINT_PATH, 10, is_report=IS_PLOT)
        manager = WorkerCheckpointManager(CHECKPOINT_PATH, is_report=IS_PLOT)
        manager.remove()
        for ffile in REMOVE_FILES:
            if os.path.exists(ffile):
                os.remove(ffile)
        
    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertTrue(os.path.exists(REFERENCE_SERIALIZATION_PATH))
        self.assertTrue(os.path.exists(TARGET_SERIALIZATION_PATH))
        self.assertGreater(len(self.finder.reference_networks), 0)

    @util.timeit
    def testFindOneProcess(self):
        if IGNORE_TEST:
            return
        df = self.finder.parallelFind(is_report=IS_PLOT, is_initialize=True, num_worker=1)
        self.assertEqual(len(df), NUM_NETWORK**2)
        prune_df = WorkerCheckpointManager.prune(df)[0]
        self.assertGreaterEqual(len(prune_df), NUM_NETWORK)
    
    @util.timeit
    def testFindOneProcessWithCheckpoint(self):
        if IGNORE_TEST:
            return
        _ = self.finder.parallelFind(is_report=IS_PLOT, is_initialize=True, num_worker=1)
        df = self.finder.parallelFind(is_report=IS_PLOT, is_initialize=False, num_worker=1)
        self.assertEqual(len(df), NUM_NETWORK**2)
        prune_df = WorkerCheckpointManager.prune(df)[0]
        self.assertGreaterEqual(len(prune_df), NUM_NETWORK)
    
    @util.timeit
    def testFindManyProcess(self):
        if IGNORE_TEST:
            return
        df = self.finder.parallelFind(is_report=IS_PLOT, is_initialize=True, num_worker=2,
              max_num_assignment=1e9)
        self.assertEqual(len(df), NUM_NETWORK**2)
        prune_df = WorkerCheckpointManager.prune(df).pruned_df
        self.assertEqual(len(prune_df), NUM_NETWORK)
        #
        df = self.finder.parallelFind(is_report=IS_PLOT, is_initialize=True, num_worker=-1,
              max_num_assignment=1e1)
        self.assertEqual(len(df), NUM_NETWORK**2)
        prune2_df = WorkerCheckpointManager.prune(df).pruned_df
        self.assertLessEqual(len(prune2_df), len(prune_df))

    @util.timeit
    def testFindScale(self):
        if IGNORE_TEST:
            return
        NUM_REFERENCE_MODEL = 10
        NUM_EXTRA_TARGET_MODEL = 10
        NETWORK_SIZE = 10
        fill_size = 10
        self.remove()
        # Construct the models
        reference_models = [Network.makeRandomNetworkByReactionType(NETWORK_SIZE, is_prune_species=True)
              for _ in range(NUM_REFERENCE_MODEL)]
        target_models = [r.fill(num_fill_reaction=fill_size, num_fill_species=fill_size) for r in reference_models]
        # Add extra target models
        target_models.extend([Network.makeRandomNetworkByReactionType(NETWORK_SIZE, is_prune_species=True)
              for _ in range(NUM_EXTRA_TARGET_MODEL)])
        # Do the search
        df = ParallelSubnetFinder.findFromNetworks(
              reference_models,
              target_models,
              num_worker=10,
              identity=cn.ID_STRONG,
              serialization_dir=cn.TEST_DIR,
              is_report=IS_PLOT,
              is_initialize=True,
              checkpoint_path=CHECKPOINT_PATH,
              reference_serialization_filename=REFERENCE_SERIALIZATION_FILENAME,
              target_serialization_filename=TARGET_SERIALIZATION_FILENAME)
        prune_result = WorkerCheckpointManager.prune(df)
        self.assertLessEqual(len(prune_result.pruned_df), NUM_REFERENCE_MODEL)
    
    @util.timeit
    def testDirectoryFind(self):
        if IGNORE_TEST:
            return
        ffiles = [f for f in os.listdir(BIOMODELS_DIR) if f.endswith(".xml")]
        df = ParallelSubnetFinder.directoryFind(BIOMODELS_DIR,
              BIOMODELS_DIR, identity=cn.ID_WEAK,
              is_report=IS_PLOT, is_initialize=True, 
              checkpoint_path=CHECKPOINT_PATH,
              max_num_assignment=1e2)
        count = len(df[df[cn.FINDER_REFERENCE_NAME] == df[cn.FINDER_TARGET_NAME]])
        self.assertEqual(count, len(ffiles))

    @util.timeit
    def testBiomodelsFindSimple(self):
        if IGNORE_TEST:
            return
        self.remove()
        max_num_network = 2
        df = ParallelSubnetFinder.biomodelsFind(
              max_num_reference_network=max_num_network,
              max_num_target_network=max_num_network,
              reference_network_size=5,
              serialization_dir=cn.TEST_DIR,
              reference_serialization_filename=REFERENCE_SERIALIZATION_FILENAME,
              target_serialization_filename=TARGET_SERIALIZATION_FILENAME,
              is_report=IS_PLOT,
              checkpoint_path=CHECKPOINT_PATH,
              identity=cn.ID_STRONG)
        prune_result = WorkerCheckpointManager.prune(df)
        self.assertGreaterEqual(len(prune_result.full_df), max_num_network**2)


if __name__ == '__main__':
    unittest.main(failfast=True)


"""
Key structures in pySubnetSB

A NamedMatrix is a two dimensional numpy array with row and column names.
A Network is a representation of a chemical reaction network that includes a NamedMatrix for reactants and products.
A Constraint describes numerical or categorical charateristics of a network that are organized as one or more NamedMatrix.
"""
__version__ = "1.0.2"



'''Worker for running SubnetFinder in parallel.'''

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.checkpoint_manager import CheckpointManager # type: ignore

import collections
import json
import numpy as np
import pandas as pd  # type: ignore
from typing import Tuple, Optional

PruneResult = collections.namedtuple("PruneResult", ["full_df", "pruned_df", "num_truncated",
      "reference_names"])


class WorkerCheckpointManager(CheckpointManager):
    RecoverResult = collections.namedtuple("RecoverResult", ["full_df", "pruned_df", "processeds"])
    MergedCheckpointResult = collections.namedtuple("MergedCheckpointResult",
           ["num_reference_network", "merged_checkpoint_manager", "dataframe",
            "worker_checkpoint_managers"])

    # Specialization of CheckpointManager for executeTask to checkpoint worker results

    def __init__(self, worker_checkpoint_path:str, is_report:bool=True, is_initialize:bool=False)->None:
        """
        Args:
            subnet_finder (SubnetFinder): SubnetFinder instance
            worker_checkpoint_path (str): Path to the CSV file for the worker checkpoint
            is_report (bool): If True, reports progress
        """
        super().__init__(worker_checkpoint_path, is_report=is_report, is_initialize=is_initialize)

    def recover(self)->RecoverResult:
        """
        Recovers a previously saved DataFrame. The recovered dataframe deletes entries with model strings that are null.

        Returns:
            pd.DataFrame: DataFrame of the checkpoint
            pd.DataFrame: DataFrame of the checkpoint stripped of null entries
            np.ndarray: List of processed reference networks
        """
        df = super().recover()
        is_success = False
        if len(df) > 0:
            try:
                full_df = pd.read_csv(self.path)
                is_success = True
            except:
                pass
        if is_success:
            prune_result = WorkerCheckpointManager.prune(full_df)
            pruned_df = prune_result.pruned_df
            reference_names = prune_result.reference_names
            # Convert the JSON string to a dictionary
            if len(pruned_df) > 0:
                pruned_df.loc[:, cn.FINDER_NAME_DCT] = pruned_df[cn.FINDER_NAME_DCT].apply(lambda x: json.loads(x))
        else:
            full_df = pd.DataFrame()
            pruned_df = pd.DataFrame()
            reference_names = []
        self._print(f"Recovering {len(reference_names)} processed models from {self.path}")
        return self.RecoverResult(full_df=full_df, pruned_df=pruned_df, processeds=reference_names)

    @staticmethod 
    def prune(df:pd.DataFrame)->PruneResult:
        """
        Prunes a DataFrame to include only rows where the reference network is not the null string.

        Args:
            df (pd.DataFrame): Table of matching networks
                reference_model (str): Reference model name
                target_model (str): Target model name
                reference_network (str): may be the null string
                target_network (str): may be the null string

        Returns: PruneResult
            pd.DataFrame: Pruned DataFrame
            list: List of reference networks that were pruned
            num_truncated (int): Number of entries where search is truncated
            reference_names (list): Names of reference networks processed
        """
        is_null = df[cn.FINDER_REFERENCE_NETWORK].isnull()
        is_null_str = df[cn.FINDER_REFERENCE_NETWORK] == cn.NULL_STR
        not_sel = is_null | is_null_str
        reference_names = list(set(df[not_sel][cn.FINDER_REFERENCE_NAME].values))
        num_truncated = np.sum(df[cn.FINDER_IS_TRUNCATED])
        prune_result = PruneResult(full_df=df, pruned_df=df[~not_sel], num_truncated=num_truncated,
              reference_names=reference_names)
        return prune_result

    @staticmethod
    def makeWorkerCheckpointPath(outpath_base, worker_idx:int)->str:
        """
        Constructs the checkpoint path for a worker

        Args:
            outpath_base (str): Path for the base file
            worker_idx (int): Index of the worker

        Returns:
            str: Path to the checkpoint
        """
        splits = outpath_base.split(".")
        outpath_pat = splits[0] + "_%d." + splits[1]
        return outpath_pat % worker_idx
    
    @classmethod
    def merge(cls, base_checkpoint_path:str, num_worker:int,
        merged_checkpoint_result:Optional[MergedCheckpointResult]=None,
        is_report:bool=True)->MergedCheckpointResult:
        """
        Merges the checkpoints from checkpoint managers. Assumes that worker checkpoint files are named
        with the pattern base_checkpoint_path_%d.csv.

        Args:
            outpath_base (str): Base path for the checkpoint files
            num_worker (int): Number of workers
            merged_checkpoint_result (Optional[MergedCheckpointResult]): Previous Merged checkpoint result
            is_report (bool): If True, reports progress

        Returns: MergedCheckpointResult
            int: Number of merged entries
            CheckpointManager: Checkpoint manager for merged checkpoint
            pd.DataFrame
        """
        if merged_checkpoint_result is None:
            merged_checkpoint_manager = WorkerCheckpointManager(base_checkpoint_path, is_report=is_report)
            worker_checkpoint_managers = [WorkerCheckpointManager(
                    WorkerCheckpointManager.makeWorkerCheckpointPath(base_checkpoint_path, i),
                    is_report=is_report, is_initialize=False)
                    for i in range(num_worker)]
        else:
            merged_checkpoint_manager = merged_checkpoint_result.merged_checkpoint_manager
            worker_checkpoint_managers = merged_checkpoint_result.worker_checkpoint_managers
        recovers = [m.recover().full_df for m in worker_checkpoint_managers]
        is_initialized = False
        if len(recovers) > 0:
            full_df = pd.concat(recovers, ignore_index=True)
            if len(full_df) > 0:
                merged_checkpoint_manager.checkpoint(full_df)
                num_reference_network = len(full_df[cn.FINDER_REFERENCE_NAME].unique())
                is_initialized = True
        if not is_initialized:
            full_df = pd.DataFrame()
            num_reference_network = 0
        result = cls.MergedCheckpointResult(
                num_reference_network=num_reference_network,
                merged_checkpoint_manager=merged_checkpoint_manager,
                worker_checkpoint_managers=worker_checkpoint_managers,
                dataframe=full_df)           
        return result
    
    @classmethod
    def deleteWorkerCheckpoints(cls, base_checkpoint_path:str, num_worker:int, is_report:bool=True)->None:
        """
        Deletes the worker checkpoints

        Args:
            base_checkpoint_path (str): Base path for the checkpoint files
            num_worker (int): Number of workers
        """
        worker_checkpoint_managers = [cls(cls.makeWorkerCheckpointPath(base_checkpoint_path, i),
              is_report=is_report)
              for i in range(num_worker)]
        [m.remove() for m in worker_checkpoint_managers]